import uuid as uniqueid
from configparser import ConfigParser
from datetime import datetime
from typing import Type

import bcrypt
from sqlalchemy import create_engine, Engine
from sqlalchemy.orm import Session
from gnatwriter.controllers import ActivityController, AuthorController, BibliographyController, ChapterController, \
    CharacterController, EventController, ImageController, LinkController, AssistantController, \
    LocationController, NoteController, SceneController, StoryController, SubmissionController, UserController, \
    OllamaModelController, ExportController
from gnatwriter.models import Base, User


def hash_password(password: str) -> str:
    """Hash a password, return hashed password"""

    if password == '':
        raise ValueError('The password cannot be empty.')

    if len(password) < 8:
        raise ValueError('The password must be at least 8 characters.')

    if len(password) > 24:
        raise ValueError('The password cannot be more than 24 characters.')

    return bcrypt.hashpw(
        password.encode('utf8'), bcrypt.gensalt(rounds=12)
    ).decode('utf8')


def verify_password(password: str, hashed_password: str) -> bool:
    """Verify a password, return true if verified, false if not"""

    return bcrypt.checkpw(
        password.encode('utf8'), hashed_password.encode('utf8')
    )


class GnatWriter:
    """Main application class for GnatWriter

    This class is a Singleton class that is used to manage the application. It
    is responsible for creating the database engine and session, and
    initializing the controllers.

    Attributes:
    -----------
    _instance: GnatWriter
        The instance of the GnatWriter class
    _engine: Engine
        The database engine
    _session: Session
        The database session
    _owner: User
        The owner of the application
    _config: ConfigParser
        The configuration parser
    database_type: str
        The type of database being used
    path_to_config: str
        The path to the configuration file

    Methods:
    --------
    __new__(cls, path_to_config: str) -> GnatWriter
        Enforce Singleton pattern
    __init__(self, path_to_config: str)
        Initialize the application
    __call__(self, *args, **kwargs) -> Controller
        Return the controller requested
    __str__(self) -> str
        Return a string representation of the application
    __repr__(self) -> str
        Return a string representation of the application
    """

    _instance: "GnatWriter" = None
    _engine: "Engine" = None
    _session: "Session" = None
    _owner: Type[User] = None
    _config: ConfigParser = None
    database_type: str = None
    path_to_config: str = None
    
    def __new__(cls, path_to_config: str):
        """Enforce Singleton pattern"""

        if cls._instance is None:
            cls._instance = super(GnatWriter, cls).__new__(cls)

        return cls._instance

    def __init__(self, path_to_config: str):

        self.path_to_config = path_to_config
        self._config = ConfigParser()
        self._config.read(self.path_to_config)
        self.database_type = self._config.get("default_database", "type")

        if self.database_type == "sqlite":
            database = self._config.get("default_database", "database")
            self._engine = create_engine(f"sqlite:///{database}")
        elif self.database_type == "postgresql":
            user = self._config.get("default_database", "user")
            password = self._config.get("default_database", "password")
            host = self._config.get("default_database", "host")
            port = self._config.get("default_database", "port")
            database = self._config.get("default_database", "database")
            self._engine = create_engine(
                f"postgresql+psycopg://{user}:{password}@{host}:{port}/{database}"
            )
        elif self.database_type == "mysql":
            user = self._config.get("default_database", "user")
            password = self._config.get("default_database", "password")
            host = self._config.get("default_database", "host")
            port = self._config.get("default_database", "port")
            database = self._config.get("default_database", "database")
            self._engine = create_engine(
                f"mysql+mysqlconnector://{user}:{password}@{host}:{port}/{database}"
            )

        Base.metadata.create_all(self._engine)
        self._session = Session(bind=self._engine, expire_on_commit=False)
        self._owner = self._session.query(User).filter(
            User.username == "gnatwriter"
        ).first()

        if self._owner is None:
            new_uuid = str(uniqueid.uuid4())
            username = "gnatwriter"
            password = hash_password("password")
            email = "gnatwriter@example.com"
            is_active = True
            is_banned = False
            created = datetime.now()
            modified = created
            user = User(
                uuid=new_uuid, username=username, password=password,
                email=email, is_active=is_active, is_banned=is_banned,
                created=created, modified=modified
            )
            self._session.add(user)
            self._session.commit()

        self._controllers = {
            "activity": ActivityController(
                self._config, self._session, self._owner
            ),
            "assistant": AssistantController(
                self._config, self._session, self._owner
            ),
            "author": AuthorController(
                self._config, self._session, self._owner
            ),
            "bibliography": BibliographyController(
                self._config, self._session, self._owner
            ),
            "chapter": ChapterController(
                self._config, self._session, self._owner
            ),
            "character": CharacterController(
                self._config, self._session, self._owner
            ),
            "event": EventController(
                self._config, self._session, self._owner
            ),
            "export": ExportController(
                self._config, self._session, self._owner
            ),
            "image": ImageController(
                self._config, self._session, self._owner
            ),
            "link": LinkController(
                self._config, self._session, self._owner
            ),
            "location": LocationController(
                self._config, self._session, self._owner
            ),
            "note": NoteController(
                self._config, self._session, self._owner
            ),
            "ollama-model": OllamaModelController(
                self._config, self._session, self._owner
            ),
            "scene": SceneController(
                self._config, self._session, self._owner
            ),
            "story": StoryController(
                self._config, self._session, self._owner
            ),
            "submission": SubmissionController(
                self._config, self._session, self._owner
            ),
            "user": UserController(
                self._config, self._session, self._owner
            )
        }

    def __call__(self, *args, **kwargs):
        return self._controllers[args[0]]

    def __str__(self):
        return "GnatWriter Application"

    def __repr__(self):
        return f"{self.__class__.__name__}()"



from enum import Enum


class CharacterRelationshipTypes(Enum):
    """The CharacterRelationshipTypes class represents the types of relationships between characters.

    Attributes
    ----------
        FAMILY: str
            The relationship is familial
        PERSONAL: str
            The relationship is personal
        ROMANTIC: str
            The relationship is romantic
        PROFESSIONAL: str
            The relationship is professional
        OTHER: str
            The relationship is other
    """

    FAMILY = 'Family'
    PERSONAL = 'Personal'
    ROMANTIC = 'Romantic'
    PROFESSIONAL = 'Professional'
    OTHER = 'Other'



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy import func
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Author, Activity, Story


class AuthorController(BaseController):
    """Author controller encapsulates author management functionality

    Attributes
    ----------
    _instance : AuthorController
        The instance of the author controller
    _config : ConfigParser
        The configuration parser
    _session : Session
        The database session object
    _owner : User
        The current user of the controller

    Methods
    -------
    create_author(name: str, initials: str, is_pseudonym: bool)
        Create a new author
    update_author(author_id: int, name: str, initials: str)
        Update an author
    delete_author_by_id(author_id: int)
        Delete an author
    get_author_by_id(author_id: int)
        Get an author by id
    get_author_by_name(name: str)
        Get an author by name
    get_author_count()
        Get author count associated with a user
    get_all_authors()
        Get all authors associated with a user
    get_authors_page(page: int, per_page: int)
        Get a single page of authors from the database associated with a user
    get_authors_by_story_id(story_id: int)
        Get all authors associated with a story
    search_authors(search: str)
        Search for authors by name
    """

    def __init__(
            self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_author(
        self, name: str, initials: str = None, is_pseudonym: bool = False
    ) -> Author:
        """Create a new author

        Parameters
        ----------
        name : str
            The name of the new author
        initials : str
            The initials of the new author, optional
        is_pseudonym : bool
            Whether the author is a pseudonym, optional, default is False, suggesting it is the user's actual name

        Returns
        -------
        Author
            The new author object
        """

        with self._session as session:

            try:

                name_exists = session.query(Author).filter(
                    Author.name == name,
                    Author.user_id == self._owner.id
                ).first()

                if name_exists:
                    raise Exception('That author already exists.')

                created = datetime.now()
                modified = created

                author = Author(
                    user_id=self._owner.id, name=name, initials=initials,
                    is_pseudonym=is_pseudonym, created=created,
                    modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Author {author.name} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(author)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return author

    def update_author(
        self, author_id: int, name: str, initials: str = None,
        is_pseudonym: bool = False
    ) -> Type[Author]:
        """Update an author

        Parameters
        ----------
        author_id : int
            The id of the author to update
        name : str
            The name of the author
        initials : str
            The initials of the author, optional
        is_pseudonym : bool
            Whether the author is a pseudonym, optional, default is False

        Returns
        -------
        Author
            The updated author object
        """

        with self._session as session:

            try:

                author = session.query(Author).filter(
                    Author.id == author_id,
                    Author.user_id == self._owner.id
                ).first()

                if not author:
                    raise ValueError('Author not found.')

                author.name = name
                author.initials = initials
                author.is_pseudonym = is_pseudonym
                author.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Author {author.name} \
                    updated by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return author

    def delete_author_by_id(self, author_id: int) -> bool:
        """Delete an author

        Parameters
        ----------
        author_id : int
            The id of the author to delete

        Returns
        -------
        bool
            True if the author was deleted, False if not
        """

        with self._session as session:

            try:

                author = session.query(Author).filter(
                    Author.id == author_id,
                    Author.user_id == self._owner.id
                ).first()

                if not author:
                    raise ValueError('Author not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Author {author.name} \
                    deleted by {self._owner.username}', created=datetime.now()
                )

                session.delete(author)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_author_by_id(self, author_id: int) -> Type[Author] | None:
        """Get an author by id

        Parameters
        ----------
        author_id : int
            The id of the author to get

        Returns
        -------
        Author | None
            The author object or None if not found
        """

        with self._session as session:

            author = session.query(Author).filter(
                Author.id == author_id,
                Author.user_id == self._owner.id
            ).first()

            return author if author else None

    def get_author_by_name(self, name: str) -> Type[Author] | None:
        """Get an author by name

        Parameters
        ----------
        name : str
            The name of the author to get

        Returns
        -------
        Author | None
            The author object or None if not found
        """

        with self._session as session:

            author = session.query(Author).filter(
                Author.name == name,
                Author.user_id == self._owner.id
            ).first()

            return author if author else None

    def get_author_count(self) -> int:
        """Get author count associated with a user

        Returns
        -------
        int
            The number of authors
        """

        with self._session as session:

            return session.query(func.count(Author.id)).filter(
                Author.user_id == self._owner.id
            ).scalar()

    def get_all_authors(self) -> List[Type[Author]]:
        """Get all authors associated with a user

        Returns
        -------
        list
            A list of author objects
        """

        with self._session as session:

            return session.query(Author).filter(
                Author.user_id == self._owner.id
            ).all()

    def get_authors_page(self, page: int, per_page: int) -> list:
        """Get a single page of authors from the database associated with a user

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of author objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Author).filter(
                Author.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def get_authors_by_story_id(
        self, story_id: int
    ) -> List[Type[Author]] | None:
        """Get all authors associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list | None
            A list of author objects or None if not found
        """

        with self._session as session:

            story = session.query(Story).filter(
                Story.id == story_id,
                Story.user_id == self._owner.id
            ).first()

            authors = []

            if story:
                for author_story in story.authors:
                    authors.append(author_story.author)

            return authors if authors else None

    def search_authors(self, search: str) -> List[Type[Author]]:
        """Search for authors by name

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of author objects
        """

        with self._session as session:

            return session.query(Author).filter(
                Author.name.like(f'%{search}%'),
                Author.user_id == self._owner.id
            ).all()



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy import or_, func
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Location, Activity, Image, ImageLocation, Link, LinkLocation, Note, LocationNote


class LocationController(BaseController):
    """Location controller encapsulates location management functionality

    Attributes
    ----------
    _instance : LocationController
        The instance of the location controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the location controller
    _session : Session
        The database session

    Methods
    -------
    create_location(title: str, description: str, address: str, city: str, state: str, country: str, zip_code: str, \
                    latitude: float, longitude: float)
        Create a new location
    update_location(location_id: int, title: str, description: str, address: str, city: str, state: str, country: str, \
                    zip_code: str, latitude: float, longitude: float)
        Update a location
    delete_location(location_id: int)
        Delete a location
    get_all_locations()
        Get all locations associated with a user
    get_all_locations_page(page: int, per_page: int)
        Get a single page of locations associated with a user from the database
    search_locations_by_title_and_description(search: str)
        Search for locations by title and description
    search_locations_by_address(search: str)
        Search for locations by address
    search_locations_by_city(search: str)
        Search for locations by city
    search_locations_by_state(search: str)
        Search for locations by state
    search_locations_by_country(search: str)
        Search for locations by country
    search_locations_by_zip_code(search: str)
        Search for locations by zip code
    append_images_to_location(location_id: int, images: list)
        Append images to a location
    get_images_by_location_id(location_id: int)
        Get all images associated with a location
    get_images_page_by_location_id(location_id: int, page: int, per_page: int)
        Get a single page of images associated with a location from the database
    append_links_to_location(location_id: int, links: list)
        Append links to a location
    get_links_by_location_id(location_id: int)
        Get all links associated with a location
    get_links_page_by_location_id(location_id: int, page: int, per_page: int)
        Get a single page of links associated with a location from the database
    append_notes_to_location(location_id: int, notes: list)
        Append notes to a location
    get_notes_by_location_id(location_id: int)
        Get all notes associated with a location
    get_notes_page_by_location_id(location_id: int, page: int, per_page: int)
        Get a single page of notes associated with a location from the database
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_location(
        self, title: str, description: str = None, address: str = None,
        city: str = None, state: str = None, country: str = None,
        zip_code: str = None, latitude: float = None, longitude: float = None
    ) -> Location:
        """Create a new location

        Parameters
        ----------
        title : str
            The title of the location
        description : str
            The description of the location, optional
        address : str
            The address of the location, optional
        city : str
            The city of the location, optional
        state : str
            The state of the location, optional
        country : str
            The country of the location, optional
        zip_code : str
            The zip code of the location, optional
        latitude : float
            The latitude of the location, optional
        longitude : float
            The longitude of the location, optional

        Returns
        -------
        Location
            The new location object
        """

        with self._session as session:
            try:
                created = datetime.now()
                modified = created

                location = Location(
                    user_id=self._owner.id, title=title,
                    description=description, address=address, city=city,
                    state=state, country=country, zip_code=zip_code,
                    latitude=latitude, longitude=longitude, created=created,
                    modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Location {location.id} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(location)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return location

    def update_location(
        self, location_id: int, title: str, description: str = None,
        address: str = None, city: str = None, state: str = None,
        country: str = None, zip_code: str = None, latitude: float = None,
        longitude: float = None
    ) -> Type[Location]:
        """Update a location

        Parameters
        ----------
        location_id : int
            The id of the location
        title : str
            The title of the location
        description : str
            The description of the location
        address : str
            The address of the location
        city : str
            The city of the location
        state : str
            The state of the location
        country : str
            The country of the location
        zip_code : str
            The zip code of the location
        latitude : float
            The latitude of the location
        longitude : float
            The longitude of the location

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                location = session.query(Location).filter(
                    Location.id == location_id,
                    Location.user_id == self._owner.id
                ).first()

                if not location:
                    raise ValueError('Location not found.')

                location.title = title
                location.description = description
                location.address = address
                location.city = city
                location.state = state
                location.country = country
                location.zip_code = zip_code
                location.latitude = latitude
                location.longitude = longitude
                location.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Location {location.id} \
                    updated by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return location

    def delete_location(self, location_id: int) -> bool:
        """Delete a location

        Parameters
        ----------
        location_id : int
            The id of the location

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                location = session.query(Location).filter(
                    Location.id == location_id,
                    Location.user_id == self._owner.id
                ).first()

                if not location:
                    raise ValueError('Location not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Location {location.id} \
                    deleted by {self._owner.username}', created=datetime.now()
                )

                session.delete(location)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_all_locations(self) -> List[Type[Location]]:
        """Get all locations associated with an owner

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            return session.query(Location).filter(
                Location.user_id == self._owner.id
            ).all()

    def get_all_locations_page(
        self, page: int, per_page: int
    ) -> List[Type[Location]]:
        """Get a single page of locations associated with an owner from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Location).filter(
                Location.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def search_locations_by_title_and_description(
        self, search: str
    ) -> List[Type[Location]]:
        """Search for locations by title and description

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            return session.query(Location).filter(
                Location.user_id == self._owner.id,
                or_(
                    Location.title.like(f'%{search}%'),
                    Location.description.like(f'%{search}%')
                )
            ).all()

    def search_locations_by_address(self, search: str) -> List[Type[Location]]:
        """Search for locations by address

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            return session.query(Location).filter(
                Location.address.like(f'%{search}%'),
                Location.user_id == self._owner.id
            ).all()

    def search_locations_by_city(self, search: str) -> List[Type[Location]]:
        """Search for locations by city

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            return session.query(Location).filter(
                Location.city.like(f'%{search}%'),
                Location.user_id == self._owner.id
            ).all()

    def search_locations_by_state(self, search: str) -> List[Type[Location]]:
        """Search for locations by state

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            return session.query(Location).filter(
                Location.state.like(f'%{search}%'),
                Location.user_id == self._owner.id
            ).all()

    def search_locations_by_country(self, search: str) -> List[Type[Location]]:
        """Search for locations by country

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            return session.query(Location).filter(
                Location.country.like(f'%{search}%'),
                Location.user_id == self._owner.id
            ).all()

    def search_locations_by_zip_code(self, search: str) -> List[Type[Location]]:
        """Search for locations by zip code

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of location objects
        """

        with self._session as session:
            return session.query(Location).filter(
                Location.zip_code.like(f'%{search}%'),
                Location.user_id == self._owner.id
            ).all()

    def append_images_to_location(
        self, location_id: int, image_ids: list
    ) -> Type[Location]:
        """Append images to a location

        Parameters
        ----------
        location_id : int
            The id of the location
        image_ids : list
            A list of image ids

        Returns
        -------
        Location
            The updated location object
        """

        with self._session as session:
            try:
                location = session.query(Location).filter(
                    Location.id == location_id,
                    Location.user_id == self._owner.id
                ).first()

                if not location:
                    raise ValueError('Location not found.')

                for image_id in image_ids:
                    image = session.query(Image).filter(
                        Image.id == image_id,
                        Image.user_id == self._owner.id
                    ).first()

                    if not image:
                        raise ValueError('Image not found.')

                    position = session.query(
                        func.max(ImageLocation.position)
                    ).filter(
                        ImageLocation.location_id == location_id
                    ).scalar()
                    position = position + 1 if position else 1
                    is_default = False
                    created = datetime.now()
                    modified = created

                    image_location = ImageLocation(
                        user_id=self._owner.id, location_id=location_id,
                        image_id=image_id, position=position,
                        is_default=is_default, created=created,
                        modified=modified
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Image \
                        {image.caption[:50]} associated with location \
                        {location.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(image_location)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return location

    def get_images_by_location_id(self, location_id: int) -> List[Type[Image]]:
        """Get all images associated with a location

        The images should be returned in the order determined by the position field in the ImageLocation table.

        Parameters
        ----------
        location_id : int
            The id of the location

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:

            for image_location in session.query(ImageLocation).filter(
                ImageLocation.location_id == location_id,
                ImageLocation.user_id == self._owner.id
            ).order_by(ImageLocation.position).all():

                yield session.query(Image).filter(
                    Image.id == image_location.image_id,
                    Image.user_id == self._owner.id
                ).first()

    def get_images_page_by_location_id(
        self, location_id: int, page: int, per_page: int
    ) -> List[Type[Image]]:
        """Get a single page of images associated with a location from the database

        Parameters
        ----------
        location_id : int
            The id of the location
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(ImageLocation).filter(
                ImageLocation.location_id == location_id,
                ImageLocation.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def append_links_to_location(
        self, location_id: int, link_ids: list
    ) -> Type[Location]:
        """Append links to a location

        Parameters
        ----------
        location_id : int
            The id of the location
        link_ids : list
            A list of link ids

        Returns
        -------
        Location
            The updated location object
        """

        with self._session as session:
            try:
                location = session.query(Location).filter(
                    Location.id == location_id,
                    Location.user_id == self._owner.id
                ).first()

                if not location:
                    raise ValueError('Location not found.')

                for link_id in link_ids:
                    link = session.query(Link).filter(
                        Link.id == link_id,
                        Link.user_id == self._owner.id
                    ).first()

                    if not link:
                        raise ValueError('Link not found.')

                    link_location = LinkLocation(
                        user_id=self._owner.id, location_id=location_id,
                        link_id=link_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Link \
                        {link.title[:50]} associated with location \
                        {location.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(link_location)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return location

    def get_links_by_location_id(self, location_id: int) -> List[Type[Link]]:
        """Get all links associated with a location

        Parameters
        ----------
        location_id : int
            The id of the location

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            for link_location in session.query(LinkLocation).filter(
                LinkLocation.location_id == location_id,
                LinkLocation.user_id == self._owner.id
            ).all():
                yield session.query(Link).filter(
                    Link.id == link_location.link_id,
                    Link.user_id == self._owner.id
                ).first()

    def get_links_page_by_location_id(
        self, location_id: int, page: int, per_page: int
    ) -> List[Type[Link]]:
        """Get a single page of links associated with a location from the database

        Parameters
        ----------
        location_id : int
            The id of the location
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(LinkLocation).filter(
                LinkLocation.location_id == location_id,
                LinkLocation.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def append_notes_to_location(
        self, location_id: int, note_ids: list
    ) -> Type[Location]:
        """Append notes to a location

        Parameters
        ----------
        location_id : int
            The id of the location
        note_ids : list
            A list of note ids

        Returns
        -------
        Location
            The updated location object
        """

        with self._session as session:
            try:
                location = session.query(Location).filter(
                    Location.id == location_id,
                    Location.user_id == self._owner.id
                ).first()

                if not location:
                    raise ValueError('Location not found.')

                for note_id in note_ids:
                    note = session.query(Note).filter(
                        Note.id == note_id,
                        Note.user_id == self._owner.id
                    ).first()

                    if not note:
                        raise ValueError('Note not found.')

                    location_note = LocationNote(
                        user_id=self._owner.id, location_id=location_id,
                        note_id=note_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Note \
                        {note.title[:50]} associated with location \
                        {location.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(location_note)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return location

    def get_notes_by_location_id(self, location_id: int) -> List[Type[Note]]:
        """Get all notes associated with a location

        Parameters
        ----------
        location_id : int
            The id of the location

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            for location_note in session.query(LocationNote).filter(
                LocationNote.location_id == location_id,
                LocationNote.user_id == self._owner.id
            ).all():
                yield session.query(Note).filter(
                    Note.id == location_note.note_id,
                    Note.user_id == self._owner.id
                ).first()

    def get_notes_page_by_location_id(
        self, location_id: int, page: int, per_page: int
    ) -> List[Type[Note]]:
        """Get a single page of notes associated with a location from the database

        Parameters
        ----------
        location_id : int
            The id of the location
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(LocationNote).filter(
                LocationNote.location_id == location_id,
                LocationNote.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, Boolean, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Image, Location, Base


class ImageLocation(Base):
    """The ImageLocation class represents the relationship between an image and a location.

    Attributes
    ----------
        user_id: int
            The id of the owner of this entry
        image_id: int
            The image's id
        location_id: int
            The location's id
        position: int
            The position of the image in the location's gallery
        is_default: bool
            Whether the image is the default image for the location
        created: str
            The creation datetime of the link between the Image and the Location
        modified: str
            The last modification datetime of the link between the Image and the Location
        user: User
            The user who owns this entry
        image: Image
            The image that the location has
        location: Location
            The location that the image belongs to

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """

    __tablename__ = 'images_locations'
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    image_id: Mapped[int] = mapped_column(Integer, ForeignKey('images.id'), primary_key=True)
    location_id: Mapped[int] = mapped_column(Integer, ForeignKey('locations.id'), primary_key=True)
    position: Mapped[int] = mapped_column(Integer, nullable=False, default=1)
    is_default: Mapped[bool] = mapped_column(Boolean, default=False)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()), onupdate=str(datetime.now()))
    user: Mapped["User"] = relationship("User")
    image: Mapped["Image"] = relationship("Image", back_populates="location")
    location: Mapped["Location"] = relationship("Location", back_populates="images")

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<ImageLocation {self.image.caption!r} - {self.location.title!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'{self.image.caption} - {self.location.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'user_id': self.user_id,
            'image_id': self.image_id,
            'location_id': self.location_id,
            'position': self.position,
            'is_default': self.is_default,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "ImageLocation":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        ImageLocation
            The unserialized relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.image_id = data.get('image_id', self.image_id)
        self.location_id = data.get('location_id', self.location_id)
        self.position = data.get('position', self.position)
        self.is_default = data.get('is_default', self.is_default)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self



import os
from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy import func, or_
from sqlalchemy.orm import Session
from gnatwriter.controllers import BaseController
from gnatwriter.models import User, Chapter, Activity, Scene, Link, ChapterLink, Note, ChapterNote


class ChapterController(BaseController):
    """Chapter controller encapsulates chapter management functionality

    Attributes
    ----------
    _instance : ChapterController
        The instance of the chapter controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the chapter controller

    Methods
    -------
    create_chapter(story_id: int, title: str, description: str)
        Create a new chapter
    update_chapter(chapter_id: int, title: str, description: str)
        Update a chapter
    delete_chapter(chapter_id: int)
        Delete a chapter
    change_chapter_position(chapter_id: int, position: int)
        Set the position of a chapter
    get_chapter_by_id(chapter_id: int)
        Get a chapter by id
    get_all_chapters()
        Get all chapters associated with a user
    get_chapters_page(page: int, per_page: int)
        Get a single page of chapters from the database associated with a user
    get_all_chapters_by_story_id(story_id: int)
        Get all chapters associated with a story
    get_chapters_page_by_story_id(story_id: int, page: int, per_page: int)
        Get a single page of chapters associated with a story from the database
    count_chapters_by_story_id(story_id: int)
        Get chapter count associated with a story
    search_chapters(search: str)
        Search for chapters by title and description belonging to a specific user
    search_chapters_by_story_id(story_id: int, search: str)
        Search for chapters by title and description belonging to a specific story
    has_scenes(chapter_id: int)
        Check if a chapter has scenes
    get_scene_by_position(chapter_id: int, position: int)
        Get a scene by position
    get_all_scenes_by_chapter_id(chapter_id: int)
        Get all scenes associated with a chapter
    append_links_to_chapter(chapter_id: int, links: list)
        Append links to a chapter
    get_links_by_chapter_id(chapter_id: int)
        Get all links associated with a chapter
    append_notes_to_chapter(chapter_id: int, notes: list)
        Append notes to a chapter
    get_notes_by_chapter_id(chapter_id: int)
        Get all notes associated with a chapter
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_chapter(
        self, story_id: int, title: str, description: str = None
    ) -> Chapter:
        """Create a new chapter

        Parameters
        ----------
        story_id : int
            The id of the story
        title : str
            The title of the chapter
        description : str
            The description of the chapter, optional

        Returns
        -------
        Chapter
            The new chapter object
        """

        with self._session as session:

            try:

                title_exists = session.query(Chapter).filter(
                    Chapter.title == title, Chapter.story_id == story_id, Chapter.user_id == self._owner.id
                ).first()

                if title_exists:
                    raise Exception('This story already has a chapter with the same title.')

                position = session.query(func.max(Chapter.position)).filter(
                    Chapter.story_id == story_id, Chapter.user_id == self._owner.id
                ).scalar()
                position = int(position) + 1 if position else 1
                created = datetime.now()
                modified = created

                chapter = Chapter(
                    user_id=self._owner.id, story_id=story_id,
                    position=position, title=title, description=description,
                    created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Chapter \
                    {chapter.title[:50]} created by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(chapter)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return chapter

    def update_chapter(
        self, chapter_id: int, title: str, description: str = None
    ) -> Type[Chapter]:
        """Update a chapter

        Parameters
        ----------
        chapter_id : int
            The id of the chapter to update
        title : str
            The title of the chapter
        description : str
            The description of the chapter, optional

        Returns
        -------
        Chapter
            The updated chapter object
        """

        with self._session as session:

            try:

                chapter = session.query(Chapter).filter(
                    Chapter.id == chapter_id,
                    Chapter.user_id == self._owner.id
                ).first()

                if not chapter:
                    raise ValueError('Chapter not found.')

                title_exists = session.query(Chapter).filter(
                    Chapter.title == title,
                    Chapter.story_id == chapter.story_id,
                    Chapter.user_id == self._owner.id
                ).first()

                if title_exists:
                    raise Exception('This story already has a chapter with the \
                    same title.')

                chapter.title = title
                chapter.description = description
                chapter.modified = str(datetime.now())

                activity = Activity(
                    user_id=self._owner.id, summary=f'Chapter {chapter.title} \
                    updated by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return chapter

    def delete_chapter(self, chapter_id: int) -> bool:
        """Delete a chapter

        Each Chapter contains an arbitrary number of Scene objects, but before those Scenes can be deleted, the links
        and notes associated with them through the LinkScene and LinkNote objects must be deleted before deleting those
        associations. Next, the scenes are deleted. Before the Chapter can be deleted, the links, notes, and joining
        associations among the ChapterLink and NoteLink objects must be deleted. The last task to perform before the
        Chapter can be deleted is to fix the hole in the position scheme its deletion is going to leave. To do that,
        we'll decrement by 1 the position for each sibling chapter sharing the same story id and having a position that
        is greater than that of the Chapter being deleted. Finally, the Chapter is deleted.

        Parameters
        ----------
        chapter_id : int
            The id of the chapter to delete

        Returns
        -------
        bool
            True if the chapter was deleted, False if not
        """

        config = ConfigParser()
        project_root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        config.read(f"{project_root}/config.cfg")

        with self._session as session:

            try:

                chapter = session.query(Chapter).filter(
                    Chapter.id == chapter_id,
                    Chapter.user_id == self._owner.id
                ).first()

                if not chapter:
                    raise ValueError('Chapter not found.')

                siblings = session.query(Chapter).filter(
                    Chapter.story_id == chapter.story_id,
                    Chapter.user_id == self._owner.id,
                    Chapter.position > chapter.position
                ).all()

                for sibling in siblings:
                    sibling.position -= 1
                    sibling.created = datetime.strptime(
                        str(sibling.created),
                        config.get("formats", "datetime")
                    )
                    sibling.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Chapter {chapter.title} \
                    deleted by {self._owner.username}', created=datetime.now()
                )

                session.delete(chapter)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def change_chapter_position(
        self, chapter_id: int, position: int
    ) -> Type[Chapter]:
        """Set the position of a chapter

        First, determine whether the new position is closer to 1 or further from 1. If closer to one, get all sibling
        chapters with positions greater than or equal to the new position but less than the current position, and
        increment those position values by 1. If the target position is further away from 1 than the current position,
        get all sibling chapters with positions greater than the current position but less than or equal to the new
        position, and decrement those position values by 1. Finally, set the position of the target chapter to the new
        position. Return the new position.

        Parameters
        ----------
        chapter_id : int
            The id of the chapter to update
        position : int
            The position of the chapter

        Returns
        -------
        Chapter
            The updated chapter object
        """

        config = ConfigParser()
        project_root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        config.read(f"{project_root}/config.cfg")

        with self._session as session:

            try:

                chapter = session.query(Chapter).filter(
                    Chapter.id == chapter_id,
                    Chapter.user_id == self._owner.id
                ).first()

                if not chapter:
                    raise ValueError('Chapter not found.')

                if position < 1:
                    raise ValueError('Position must be greater than 0.')

                highest_position = session.query(
                    func.max(Chapter.position)
                ).filter(
                    Chapter.story_id == chapter.story_id,
                    Chapter.user_id == self._owner.id
                ).scalar()

                if position > highest_position:
                    raise ValueError(f'Position must be less than or equal to \
                    {highest_position}.')

                if position == chapter.position:
                    return chapter.position

                if position < chapter.position:
                    siblings = session.query(Chapter).filter(
                        Chapter.story_id == chapter.story_id,
                        Chapter.user_id == self._owner.id,
                        Chapter.position >= position,
                        Chapter.position < chapter.position
                    ).all()

                    for sibling in siblings:
                        sibling.position += 1
                        sibling.created = datetime.strptime(
                            str(sibling.created),
                            config.get("formats", "datetime")
                        )
                        sibling.modified = datetime.now()

                else:
                    siblings = session.query(Chapter).filter(
                        Chapter.story_id == chapter.story_id,
                        Chapter.user_id == self._owner.id,
                        Chapter.position > chapter.position,
                        Chapter.position <= position
                    ).all()

                    for sibling in siblings:
                        sibling.position -= 1
                        sibling.created = datetime.strptime(
                            str(sibling.created),
                            config.get("formats", "datetime")
                        )
                        sibling.modified = datetime.now()

                chapter.position = position
                chapter.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Chapter \
                    {chapter.title[:50]} position changed by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return chapter

    def get_chapter_by_id(self, chapter_id: int) -> Type[Chapter] | None:
        """Get a chapter by id

        Parameters
        ----------
        chapter_id : int
            The id of the chapter to get

        Returns
        -------
        Chapter | None
            The chapter object or None if not found
        """

        with self._session as session:

            chapter = session.query(Chapter).filter(
                Chapter.id == chapter_id,
                Chapter.user_id == self._owner.id
            ).first()

            return chapter if chapter else None

    def get_all_chapters(self) -> List[Type[Chapter]]:
        """Get all chapters associated with a user

        Chapters are sorted by story id and position.

        Returns
        -------
        list | None
            A list of chapter objects or None if none are found
        """

        with self._session as session:

            return session.query(Chapter).filter(
                Chapter.user_id == self._owner.id
            ).order_by(Chapter.story_id, Chapter.position).all()

    def get_all_chapters_page(
        self, page: int, per_page: int
    ) -> List[Type[Chapter]]:
        """Get a single page of chapters from the database associated with a user

        Chapters are sorted by story id and position.

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list | None
            A list of chapter objects or None if none are found
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Chapter).filter(
                Chapter.user_id == self._owner.id
            ).order_by(
                Chapter.story_id, Chapter.position
            ).offset(offset).limit(per_page).all()

    def get_chapters_by_story_id(self, story_id: int) -> List[Type[Chapter]]:
        """Get all chapters associated with a story

        The returned list will be sorted by the position.

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list } None
            A list of chapter objects or None if none are found
        """

        with self._session as session:

            return session.query(Chapter).filter(
                Chapter.story_id == story_id,
                Chapter.user_id == self._owner.id
            ).order_by(Chapter.position).all()

    def get_chapters_page_by_story_id(
        self, story_id: int, page: int, per_page: int
    ) -> List[Type[Chapter]] | None:
        """Get a single page of chapters associated with a story from the database

        The returned list will be sorted by the position.

        Parameters
        ----------
        story_id : int
            The id of the story
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list | None
            A list of chapter objects or None if none are found
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Chapter).filter(
                Chapter.story_id == story_id,
                Chapter.user_id == self._owner.id
            ).order_by(
                Chapter.position
            ).offset(offset).limit(per_page).all()

    def count_chapters_by_story_id(self, story_id: int) -> int:
        """Get chapter count associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        int
            The number of chapters
        """

        with self._session as session:

            return session.query(func.count(Chapter.id)).filter(
                Chapter.story_id == story_id,
                Chapter.user_id == self._owner.id
            ).scalar()

    def search_chapters(self, search: str) -> List[Type[Chapter]]:
        """Search for chapters by title and description belonging to a specific user

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of chapter objects
        """

        with self._session as session:

            return session.query(Chapter).filter(
                or_(
                    Chapter.title.like(f'%{search}%'),
                    Chapter.description.like(f'%{search}%')
                ),
                Chapter.user_id == self._owner.id
            ).all()

    def search_chapters_by_story_id(
        self, story_id: int, search: str
    ) -> List[Type[Chapter]]:
        """Search for chapters by title and description belonging to a specific story

        Parameters
        ----------
        story_id : int
            The id of the story
        search : str
            The search string

        Returns
        -------
        list
            A list of chapter objects
        """

        with self._session as session:

            return session.query(Chapter).filter(
                or_(
                    Chapter.title.like(f'%{search}%'),
                    Chapter.description.like(f'%{search}%')
                ),
                Chapter.story_id == story_id,
                Chapter.user_id == self._owner.id
            ).all()

    def has_scenes(self, chapter_id: int) -> bool:
        """Check if a chapter has scenes

        Parameters
        ----------
        chapter_id : int
            The id of the chapter

        Returns
        -------
        bool
            True if the chapter has scenes, False if not
        """

        with self._session as session:

            return session.query(Scene).filter(
                Scene.chapter_id == chapter_id,
                Scene.user_id == self._owner.id
            ).count() > 0

    def get_scene_by_position(
        self, chapter_id: int, position: int
    ) -> Type[Scene] | None:
        """Get a scene by position

        Parameters
        ----------
        chapter_id : int
            The id of the chapter
        position : int
            The position of the scene

        Returns
        -------
        Scene | None
            The scene object or None if not found
        """

        with self._session as session:

            scene = session.query(Scene).filter(
                Scene.chapter_id == chapter_id,
                Scene.position == position,
                Scene.user_id == self._owner.id
            ).first()

            return scene if scene else None

    def get_all_scenes_by_chapter_id(
        self, chapter_id: int
    ) -> List[Type[Scene]]:
        """Get all scenes associated with a chapter

        Parameters
        ----------
        chapter_id : int
            The id of the chapter

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:

            return session.query(Scene).filter(
                Scene.chapter_id == chapter_id,
                Scene.user_id == self._owner.id
            ).order_by(Scene.position).all()

    def append_links_to_chapter(
        self, chapter_id: int, link_ids: list
    ) -> Type[Chapter]:
        """Append links to a chapter

        Parameters
        ----------
        chapter_id : int
            The id of the chapter
        link_ids : list
            A list of link ids

        Returns
        -------
        Chapter
            The updated chapter object
        """

        with self._session as session:

            try:

                chapter = session.query(Chapter).filter(
                    Chapter.id == chapter_id,
                    Chapter.user_id == self._owner.id
                ).first()

                if not chapter:
                    raise ValueError('Chapter not found.')

                for link_id in link_ids:
                    link = session.query(Link).filter(Link.id == link_id).first()

                    if not link:
                        raise ValueError('Link not found.')

                    chapter_link = ChapterLink(
                        user_id=self._owner.id, story_id=chapter.story_id,
                        chapter_id=chapter_id, link_id=link_id,
                        created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Link \
                        {link.title[:50]} associated with chapter \
                        {chapter.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(chapter_link)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return chapter

    def get_links_by_chapter_id(self, chapter_id: int) -> List[Type[Link]]:
        """Get all links associated with a chapter

        Parameters
        ----------
        chapter_id : int
            The id of the chapter

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:

            return session.query(Link).join(
                ChapterLink, Link.id == ChapterLink.link_id
            ).filter(
                ChapterLink.chapter_id == chapter_id,
                ChapterLink.user_id == self._owner.id
            ).all()

    def append_notes_to_chapter(
            self, chapter_id: int, note_ids: list
    ) -> Type[Chapter]:
        """Append notes to a chapter

        Parameters
        ----------
        chapter_id : int
            The id of the chapter
        note_ids : list
            A list of note ids

        Returns
        -------
        Chapter
            The updated chapter object
        """

        with self._session as session:

            try:

                chapter = session.query(Chapter).filter(
                    Chapter.id == chapter_id,
                    Chapter.user_id == self._owner.id
                ).first()

                if not chapter:
                    raise ValueError('Chapter not found.')

                for note_id in note_ids:
                    note = session.query(Note).filter(
                        Note.id == note_id,
                        Note.user_id == self._owner.id
                    ).first()

                    if not note:
                        raise ValueError('Note not found.')

                    chapter_note = ChapterNote(
                        user_id=self._owner.id, story_id=chapter.story_id,
                        chapter_id=chapter_id, note_id=note_id,
                        created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Note\
                        {note.title[:50]} associated with chapter \
                        {chapter.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(chapter_note)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return chapter

    def get_notes_by_chapter_id(self, chapter_id: int) -> List[Type[Note]]:
        """Get all notes associated with a chapter

        Parameters
        ----------
        chapter_id : int
            The id of the chapter

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:

            return session.query(Note).join(
                ChapterNote, Note.id == ChapterNote.note_id
            ).filter(
                ChapterNote.chapter_id == chapter_id,
                ChapterNote.user_id == self._owner.id
            ).all()



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, String, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import Bibliography, User, Base


class BibliographyAuthor(Base):
    """The BibliographyAuthor class represents an author of a reference.

    Attributes
    ----------
        id: int
            The author's id
        user_id: int
            The id of the owner of this entry
        bibliography_id: int
            The reference's id
        name: str
            The author's name
        initials: str
            The author's initials
        created: str
            The creation datetime of the link between the Bibliography reference and its Author
        user: User
            The user who owns this entry
        reference: Bibliography
            The reference that the author wrote

    Methods
    -------
        __repr__()
            Returns a string representation of the author
        __str__()
            Returns a string representation of the author
        serialize()
            Returns a dictionary representation of the author
        unserialize(data: dict)
            Updates the author's attributes with the values from the dictionary
        validate_name(name: str)
            Validates the name's length
        validate_initials(initials: str)
            Validates the initials' length
    """

    __tablename__ = 'authors_bibliographies'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    bibliography_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('bibliographies.id')
    )
    name: Mapped[str] = mapped_column(String(150), nullable=False)
    initials: Mapped[str] = mapped_column(String(10), nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship("User")
    reference: Mapped["Bibliography"] = relationship(
        "Bibliography", back_populates="authors"
    )

    def __repr__(self):
        """Returns a string representation of the author.

        Returns
        -------
        str
            A string representation of the author
        """

        return f'<BibliographyAuthor {self.bibliography.title!r} - {self.name!r}>'

    def __str__(self):
        """Returns a string representation of the author.

        Returns
        -------
        str
            A string representation of the author
        """

        return f'BibliographyAuthor: {self.bibliography.title} - {self.name}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the author.

        Returns
        -------
        dict
            A dictionary representation of the author
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'bibliography_id': self.bibliography_id,
            'name': self.name,
            'initials': self.initials,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "BibliographyAuthor":
        """Updates the author's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the author

        Returns
        -------
        BibliographyAuthor
            The updated author
        """

        self.id = data.get('id', self.id)
        self.user_id = data.get('user_id', self.user_id)
        self.bibliography_id = data.get('bibliography_id', self.bibliography_id)
        self.name = data.get('name', self.name)
        self.initials = data.get('initials', self.initials)
        self.created = data.get('created', self.created)

        return self

    @validates("name")
    def validate_name(self, key, name: str) -> str:
        """Validates the name's length.

        Parameters
        ----------
        name: str
            The author's name

        Returns
        -------
        str
            The validated name
        """

        if not name:
            raise ValueError("A author name is required.")

        if len(name) > 150:
            raise ValueError("The author's name must have no more than 150 characters.")

        return name

    @validates("initials")
    def validate_initials(self, key, initials: str) -> str:
        """Validates the initials' length.

        Parameters
        ----------
        initials: str
            The author's initials

        Returns
        -------
        str
            The validated initials
        """

        if initials and len(initials) > 10:
            raise ValueError("The author's initials must have no more than 10 characters.")

        return initials



from gnatwriter.models.Base import Base
from gnatwriter.models.Activity import Activity
from gnatwriter.models.Assistance import Assistance
from gnatwriter.models.Author import Author
from gnatwriter.models.AuthorStory import AuthorStory
from gnatwriter.models.Bibliography import Bibliography
from gnatwriter.models.BibliographyAuthor import BibliographyAuthor
from gnatwriter.models.Chapter import Chapter
from gnatwriter.models.ChapterLink import ChapterLink
from gnatwriter.models.ChapterNote import ChapterNote
from gnatwriter.models.Character import Character
from gnatwriter.models.CharacterEvent import CharacterEvent
from gnatwriter.models.CharacterImage import CharacterImage
from gnatwriter.models.CharacterLink import CharacterLink
from gnatwriter.models.CharacterNote import CharacterNote
from gnatwriter.models.CharacterRelationship import CharacterRelationship
from gnatwriter.models.CharacterRelationshipTypes import CharacterRelationshipTypes
from gnatwriter.models.CharacterStory import CharacterStory
from gnatwriter.models.CharacterTrait import CharacterTrait
from gnatwriter.models.Event import Event
from gnatwriter.models.EventLink import EventLink
from gnatwriter.models.EventNote import EventNote
from gnatwriter.models.Image import Image
from gnatwriter.models.ImageLocation import ImageLocation
from gnatwriter.models.ImageMimeTypes import ImageMimeTypes
from gnatwriter.models.Link import Link
from gnatwriter.models.LinkLocation import LinkLocation
from gnatwriter.models.LinkScene import LinkScene
from gnatwriter.models.LinkStory import LinkStory
from gnatwriter.models.Location import Location
from gnatwriter.models.LocationNote import LocationNote
from gnatwriter.models.Note import Note
from gnatwriter.models.NoteScene import NoteScene
from gnatwriter.models.NoteStory import NoteStory
from gnatwriter.models.OllamaModel import OllamaModel
from gnatwriter.models.Scene import Scene
from gnatwriter.models.Story import Story
from gnatwriter.models.Submission import Submission
from gnatwriter.models.SubmissionResultType import SubmissionResultType
from gnatwriter.models.User import User



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, Text, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, NoteStory, ChapterNote, NoteScene, CharacterNote, EventNote, LocationNote, Base


class Note(Base):
    """The Note class represents a note.

    Attributes
    ----------
        id: int
            The note's id
        user_id: int
            The id of the owner of this entry
        title: str
            The note's title
        content: str
            The note's content
        created: str
            The creation datetime of the note
        modified: str
            The last modification datetime of the note
        user: User
            The user who owns this entry
        stories: List[NoteStory]
            The stories that the note has
        chapters: List[ChapterNote]
            The chapters that the note has
        scenes: List[NoteScene]
            The scenes that the note has
        characters: List[CharacterNote]
            The characters that the note has
        events: List[EventNote]
            The events that the note has
        locations: List[LocationNote]
            The locations that the note has

    Methods
    -------
        __repr__()
            Returns a string representation of the note
        __str__()
            Returns a string representation of the note
        serialize()
            Returns a dictionary representation of the note
        unserialize(data: dict)
            Updates the note's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_content(content: str)
            Validates the content's length
    """

    __tablename__ = 'notes'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    title: Mapped[str] = mapped_column(String(250), nullable=False)
    content: Mapped[str] = mapped_column(Text, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User", back_populates="notes")
    stories: Mapped[Optional[List["NoteStory"]]] = relationship(
        "NoteStory", back_populates="note",
        cascade="all, delete, delete-orphan")
    chapters: Mapped[Optional[List["ChapterNote"]]] = relationship(
        "ChapterNote", back_populates="note",
        cascade="all, delete, delete-orphan")
    scenes: Mapped[Optional[List["NoteScene"]]] = relationship(
        "NoteScene", back_populates="note",
        cascade="all, delete, delete-orphan")
    characters: Mapped[Optional[List["CharacterNote"]]] = relationship(
        "CharacterNote", back_populates="note",
        cascade="all, delete, delete-orphan")
    events: Mapped[Optional[List["EventNote"]]] = relationship(
        "EventNote", back_populates="note",
        cascade="all, delete, delete-orphan")
    locations: Mapped[Optional[List["LocationNote"]]] = relationship(
        "LocationNote", back_populates="note",
        cascade="all, delete, delete-orphan")

    def __repr__(self):
        """Returns a string representation of the note.

        Returns
        -------
        str
            A string representation of the note
        """

        return f'<Note {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the note.

        Returns
        -------
        str
            A string representation of the note
        """

        return f'{self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the note.

        Returns
        -------
        dict
            A dictionary representation of the note
        """

        escaped_title = self.title.replace('"', '\\"').replace('/', '\\/')
        escaped_content = self.content.replace('"', '\\"').replace('/', '\\/')

        return {
            'id': self.id,
            'user_id': self.user_id,
            'title': escaped_title,
            'content': escaped_content,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Note":
        """Updates the note's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the note

        Returns
        -------
        Note
            The unserialized note
        """

        self.user_id = data.get('user_id', self.user_id)
        self.title = data.get('title', self.title)
        self.content = data.get('content', self.content)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length

        Parameters
        ----------
        title: str
            The note's title

        Returns
        -------
        str
            The validated title
        """

        if not title:
            raise ValueError("A note title is required.")

        if len(title) > 250:
            raise ValueError("The note title can have no more than 250 characters.")

        return title

    @validates("content")
    def validate_content(self, key, content: str) -> str:
        """Validates the content's length

        Parameters
        ----------
        content: str
            The note's content

        Returns
        -------
        str
            The validated content
        """

        if content and len(content) > 65535:
            raise ValueError("The note content can have no more than 65,535 characters.")

        return content



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Event, CharacterEvent, Character, Activity, Location, Link, EventLink, Note, EventNote


class EventController(BaseController):
    """Event controller encapsulates event management functionality

    Attributes
    ----------
    _instance : EventController
        The instance of the event controller
    _config : ConfigParser
        The configuration parser
    _owner : User
        The current user of the event controller
    _session : Session
        The database session

    Methods
    -------
    create_event(title: str, description: str, start_datetime: str, end_datetime: str)
        Create a new event
    update_event(event_id: int, title: str, description: str, start_datetime: str, end_datetime: str)
        Update an event
    delete_event(event_id: int)
        Delete an event
    get_all_events()
        Get all events associated with a user
    get_all_events_page(page: int, per_page: int)
        Get a single page of events associated with a user from the database
    append_characters_to_event(event_id: int, characters: list)
        Append characters to an event
    get_characters_by_event_id(event_id: int)
        Get all characters associated with an event
    append_links_to_event(event_id: int, links: list)
        Append links to an event
    get_links_by_event_id(event_id: int)
        Get all links associated with an event
    get_links_page_by_event_id(event_id: int, page: int, per_page: int)
        Get a single page of links associated with an event from the database
    append_notes_to_event(event_id: int, notes: list)
        Append notes to an event
    get_notes_by_event_id(event_id: int)
        Get all notes associated with an event
    get_notes_page_by_event_id(event_id: int, page: int, per_page: int)
        Get a single page of notes associated with an event from the database
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_event(
        self, title: str, description: str = None, start_datetime: str = None,
        end_datetime: str = None
    ) -> Event:
        """Create a new event

        Parameters
        ----------
        title : str
            The title of the event
        description : str
            The description of the event
        start_datetime : str
            The start date and time of the event
        end_datetime : str
            The end date and time of the event

        Returns
        -------
        Event
            The new event object
        """

        with self._session as session:
            try:
                created = datetime.now()
                modified = created

                event = Event(
                    user_id=self._owner.id, title=title, description=description,
                    start_datetime=start_datetime, end_datetime=end_datetime,
                    created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Event {event.title[:50]} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(event)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return event

    def update_event(
        self, event_id: int, title: str, description: str, start_datetime: str,
        end_datetime: str
    ) -> Type[Event]:
        """Update an event

        Parameters
        ----------
        event_id : int
            The id of the event
        title : str
            The title of the event
        description : str
            The description of the event
        start_datetime : str
            The start date and time of the event
        end_datetime : str
            The end date and time of the event

        Returns
        -------
        Event
            The updated event object
        """

        with self._session as session:
            try:
                event = session.query(Event).filter(Event.id == event_id).first()

                if not event:
                    raise ValueError('Event not found.')

                event.title = title
                event.description = description
                event.start_datetime = start_datetime
                event.end_datetime = end_datetime
                event.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Event {event.id} updated \
                    by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return event

    def delete_event(self, event_id: int) -> bool:
        """Delete an event

        Parameters
        ----------
        event_id : int
            The id of the event

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                event = session.query(Event).filter(
                    Event.id == event_id,
                    Event.user_id == self._owner.id
                ).first()

                if not event:
                    raise ValueError('Event not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Event {event.id} deleted \
                    by {self._owner.username}', created=datetime.now()
                )

                session.delete(event)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_all_events(self) -> List[Type[Event]]:
        """Get all events associated with a user

        Returns
        -------
        list
            A list of event objects
        """

        with self._session as session:
            return session.query(Event).filter(
                Event.user_id == self._owner.id
            ).all()

    def get_all_events_page(
        self, page: int, per_page: int
    ) -> List[Type[Event]]:
        """Get a single page of events associated with a user from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of event objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Event).filter(
                Event.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def append_characters_to_event(
        self, event_id: int, character_ids: list
    ) -> Type[Event]:
        """Append characters to an event

        Parameters
        ----------
        event_id : int
            The id of the event
        character_ids : list
            A list of character ids

        Returns
        -------
        Event
            The updated event object
        """

        with self._session as session:
            try:
                event = session.query(Event).filter(
                    Event.id == event_id,
                    Event.user_id == self._owner.id
                ).first()

                if not event:
                    raise ValueError('Event not found.')

                for character_id in character_ids:
                    character = session.query(Character).filter(
                        Character.id == character_id,
                        Character.user_id == self._owner.id
                    ).first()

                    if not character:
                        raise ValueError('Character not found.')

                    character_event = CharacterEvent(
                        user_id=self._owner.id, event_id=event_id,
                        character_id=character_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Character \
                        {character.__str__} associated with event \
                        {event.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(character_event)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return event

    def get_characters_by_event_id(
        self, event_id: int
    ) -> List[Type[Character]]:
        """Get all characters associated with an event

        Parameters
        ----------
        event_id : int
            The id of the event

        Returns
        -------
        list
            A list of character objects
        """

        with self._session as session:
            for character_event in session.query(CharacterEvent).filter(
                CharacterEvent.event_id == event_id,
                    CharacterEvent.user_id == self._owner.id
            ).all():
                yield session.query(Character).filter(
                    Character.id == character_event.character_id
                ).first()

    def get_characters_page_by_event_id(
        self, event_id: int, page: int, per_page: int
    ) -> List[Type[Character]]:
        """Get a single page of characters associated with an event from the database

        Parameters
        ----------
        event_id : int
            The id of the event
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of character objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            for character_event in session.query(CharacterEvent).filter(
                CharacterEvent.event_id == event_id,
                CharacterEvent.user_id == self._owner.id
            ).offset(offset).limit(per_page).all():
                yield session.query(Character).filter(
                    Character.id == character_event.character_id,
                    Character.user_id == self._owner.id
                ).first()

    def append_links_to_event(
        self, event_id: int, link_ids: list
    ) -> Type[Event]:
        """Append links to an event

        Parameters
        ----------
        event_id : int
            The id of the event
        link_ids : list
            A list of link ids

        Returns
        -------
        Event
            The updated event object
        """

        with self._session as session:
            try:
                event = session.query(Event).filter(
                    Event.id == event_id,
                    Event.user_id == self._owner.id
                ).first()

                if not event:
                    raise ValueError('Event not found.')

                for link_id in link_ids:
                    link = session.query(Link).filter(
                        Link.id == link_id,
                        Link.user_id == self._owner.id
                    ).first()

                    if not link:
                        raise ValueError('Link not found.')

                    event_link = EventLink(
                        user_id=self._owner.id, event_id=event_id,
                        link_id=link_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Link\
                        {link.title[:50]} associated with event \
                        {event.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(event_link)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return event

    def get_links_by_event_id(self, event_id: int) -> List[Type[Link]]:
        """Get all links associated with an event

        Parameters
        ----------
        event_id : int
            The id of the event

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            for event_link in session.query(EventLink).filter(
                EventLink.event_id == event_id,
                    EventLink.user_id == self._owner.id
            ).all():
                yield session.query(Link).filter(
                    Link.id == event_link.link_id,
                    Link.user_id == self._owner.id
                ).first()

    def get_links_page_by_event_id(
        self, event_id: int, page: int, per_page: int
    ) -> List[Type[Link]]:

        """Get a single page of links associated with an event from the database

        Parameters
        ----------
        event_id : int
            The id of the event
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(EventLink).filter(
                EventLink.event_id == event_id,
                EventLink.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def append_notes_to_event(
        self, event_id: int, note_ids: list
    ) -> Type[Event]:
        """Append notes to an event

        Parameters
        ----------
        event_id : int
            The id of the event
        note_ids : list
            A list of note ids

        Returns
        -------
        Event
            The updated event object
        """

        with self._session as session:
            try:
                event = session.query(Event).filter(
                    Event.id == event_id,
                    Event.user_id == self._owner.id
                ).first()

                if not event:
                    raise ValueError('Event not found.')

                for note_id in note_ids:
                    note = session.query(Note).filter(
                        Note.id == note_id,
                        Note.user_id == self._owner.id
                    ).first()

                    if not note:
                        raise ValueError('Note not found.')

                    event_note = EventNote(
                        user_id=self._owner.id, event_id=event_id,
                        note_id=note_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Note \
                        {note.title[:50]} associated with event \
                        {event.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(event_note)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return event

    def get_notes_by_event_id(self, event_id: int) -> List[Type[Note]]:
        """Get all notes associated with an event

        Parameters
        ----------
        event_id : int
            The id of the event

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            for event_note in session.query(EventNote).filter(
                EventNote.event_id == event_id,
                    EventNote.user_id == self._owner.id
            ).all():
                yield session.query(Note).filter(
                    Note.id == event_note.note_id,
                    Note.user_id == self._owner.id
                ).first()

    def get_notes_page_by_event_id(
        self, event_id: int, page: int, per_page: int
    ) -> List[Type[Note]]:
        """Get a single page of notes associated with an event from the database

        Parameters
        ----------
        event_id : int
            The id of the event
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(EventNote).filter(
                EventNote.event_id == event_id,
                EventNote.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, String, Date, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, Character, CharacterRelationshipTypes, Base


class CharacterRelationship(Base):
    """The CharacterRelationship class represents the relationship between two characters.

    Attributes
    ----------
        id: int
            The character relationship's id
        user_id: int
            The id of the owner of this entry
        parent_id: int
            The parent character's id
        position: int
            The position of the related character in the parent character's life
        related_id: int
            The related character's id
        relationship_type: str
            The type of relationship between the characters
        description: str
            A description of the relationship
        start_date: str
            The starting date of the relationship
        end_date: str
            The ending date of the relationship
        created: str
            The creation datetime of the relationship
        modified: str
            The last modification datetime of the relationship
        user: User
            The user who owns this entry
        parent_character: Character
            The parent character in the relationship
        related_character: Character
            The related character in the relationship

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
        validate_relationship_type(relationship_type: str)
            Validates the relationship type
        validate_description(description: str)
            Validates the description
        validate_start_date(start_date: str)
            Validates the start date
        validate_end_date(end_date: str)
            Validates the end date
    """

    __tablename__ = 'character_relationships'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    parent_id: Mapped[int] = mapped_column(Integer, ForeignKey('characters.id'))
    position: Mapped[int] = mapped_column(Integer, nullable=False, default=1)
    related_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('characters.id')
    )
    relationship_type: Mapped[str] = mapped_column(String(50), nullable=False)
    description: Mapped[str] = mapped_column(String(250), nullable=True)
    start_date: Mapped[str] = mapped_column(Date, nullable=True)
    end_date: Mapped[str] = mapped_column(Date, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User")
    parent_character: Mapped["Character"] = relationship(
        "Character", foreign_keys="CharacterRelationship.parent_id"
    )
    related_character: Mapped["Character"] = relationship(
        "Character", back_populates="character_relationships",
        lazy="joined", foreign_keys="[CharacterRelationship.related_id]"
    )

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return (f'<CharacterRelationship {self.parent_character.first_name!r} {self.parent_character.last_name!r} - '
                f'{self.description!r} - {self.related_character.first_name!r} {self.related_character.last_name!r}>')

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return (f'{self.parent_character.first_name} {self.parent_character.last_name} - '
                f'{self.description!r} - {self.related_character.first_name} {self.related_character.last_name}')

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        related_name = ""

        if self.related_character:
            related_name = self.related_character.full_name

        return {
            'user_id': self.user_id,
            'parent_id': self.parent_id,
            'position': self.position,
            'related_id': self.related_id,
            'related_name': related_name,
            'relationship_type': self.relationship_type,
            'description': self.description,
            'start_date': self.start_date,
            'end_date': self.end_date,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "CharacterRelationship":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        CharacterRelationship
            The unserialized relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.parent_id = data.get('parent_id', self.parent_id)
        self.position = data.get('position', self.position)
        self.related_id = data.get('related_id', self.related_id)
        self.relationship_type = data.get('relationship_type', self.relationship_type)
        self.description = data.get('description', self.description)
        self.start_date = data.get('start_date', self.start_date)
        self.end_date = data.get('end_date', self.end_date)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("relationship_type")
    def validate_relationship_type(self, key, relationship_type: str) -> str:
        """Validates the relationship type.

        Parameters
        ----------
        relationship_type: str
            The character relationship type

        Returns
        -------
        str
            The validated relationship type
        """

        if not relationship_type:
            raise ValueError("A character relationship type is required.")

        rtypes = [str(rt.value) for rt in CharacterRelationshipTypes]

        if relationship_type not in rtypes:
            raise ValueError("The character relationship type is invalid.")

        return relationship_type

    @validates("description")
    def validate_description(self, key, description: str) -> str:
        """Validates the description.

        Parameters
        ----------
        description: str
            The character relationship description

        Returns
        -------
        str
            The validated description
        """

        if not description:
            raise ValueError("A character relationship description is required.")

        if len(description) > 250:
            raise ValueError("The character relationship description must have no more than 250 characters.")

        return description



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Link, Activity, LinkStory


class LinkController(BaseController):
    """Link controller encapsulates link management functionality

    Attributes
    ----------
    _instance : LinkController
        The instance of the link controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the link controller
    _session : Session
        The database session

    Methods
    -------
    create_link(url: str, title: str)
        Create a new link
    update_link(link_id: int, url: str, title: str)
        Update a link
    delete_link(link_id: int)
        Delete a link
    get_link_by_id(link_id: int)
        Get a link by id
    get_all_links()
        Get all links associated with a user
    get_all_links_page(page: int, per_page: int)
        Get a single page of links associated with a user from the database
    search_links(search: str)
        Search for links by title
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_link(self, url: str, title: str) -> Link:
        """Create a new link

        Parameters
        ----------
        url : str
            The url of the link
        title : str
            The title of the link

        Returns
        -------
        Link
            The new link object
        """

        with self._session as session:
            try:
                created = datetime.now()
                modified = created

                link = Link(
                    user_id=self._owner.id, url=url, title=title,
                    created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Link {link.title[:50]} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(link)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return link

    def update_link(self, link_id: int, url: str, title: str) -> Type[Link]:
        """Update a link

        Parameters
        ----------
        link_id : int
            The id of the link
        url : str
            The url of the link
        title : str
            The title of the link

        Returns
        -------
        Link
            The updated link object
        """

        with self._session as session:
            try:
                link = session.query(Link).filter(
                    Link.id == link_id,
                    Link.user_id == self._owner.id
                ).first()

                if not link:
                    raise ValueError('Link not found.')

                link.url = url
                link.title = title
                link.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Link {link.id} updated \
                    by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return link

    def delete_link(self, link_id: int) -> bool:
        """Delete a link

        Parameters
        ----------
        link_id : int
            The id of the link

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                link = session.query(Link).filter(
                    Link.id == link_id,
                    Link.user_id == self._owner.id
                ).first()

                if not link:
                    raise ValueError('Link not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Link {link.id} deleted \
                    by {self._owner.username}', created=datetime.now()
                )

                session.delete(link)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_link_by_id(self, link_id: int) -> Type[Link] | None:
        """Get a link by id

        Parameters
        ----------
        link_id : int
            The id of the link

        Returns
        -------
        Link
            The link object
        """

        with self._session as session:
            return session.query(Link).filter(
                Link.id == link_id,
                Link.user_id == self._owner.id
            ).first()

    def get_links_by_story_id(self, story_id: int) -> List[Type[Link]]:
        """Get all links associated with a story

        The LinkStory objects are the associations between links and stories. This method returns a list of links
        associated with a story.

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            return session.query(Link).join(LinkStory).filter(
                LinkStory.story_id == story_id,
                LinkStory.user_id == self._owner.id
            ).all()

    def get_all_links(self) -> List[Type[Link]]:
        """Get all links associated with an owner

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            return session.query(Link).filter(
                Link.user_id == self._owner.id
            ).all()

    def get_all_links_page(self, page: int, per_page: int) -> List[Type[Link]]:
        """Get a single page of links associated with an owner from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Link).filter(
                Link.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def search_links(self, search: str) -> List[Type[Link]]:
        """Search for links by title

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            return session.query(Link).filter(
                Link.title.like(f'%{search}%'),
                Link.user_id == self._owner.id
            ).all()



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy import func
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Bibliography, Activity


class BibliographyController(BaseController):
    """Bibliography controller encapsulates bibliography management functionality

    Attributes
    ----------
    _config : ConfigParser
        The configuration parser
    _owner : User
        The current user of the bibliography controller
    _session : Session
        The database session

    Methods
    -------
    create_bibliography(story_id: int, title: str, pages: str, publication_date: str, other_text: str)
        Create a new bibliography
    update_bibliography(bibliography_id: int, story_id: int, title: str, pages: str, publication_date: str, \
                        other_text: str)
        Update a bibliography
    delete_bibliography(bibliography_id: int)
        Delete a bibliography
    get_bibliography_by_id(bibliography_id: int)
        Get a bibliography by id
    get_bibliography_by_title(title: str)
        Get a bibliography by title
    get_bibliography_count()
        Get bibliography count associated with a user
    get_all_bibliographies()
        Get all bibliographies associated with a user
    get_bibliographies_page(page: int, per_page: int)
        Get a single page of bibliographies from the database associated with a user
    get_bibliographies_by_story_id(story_id: int)
        Get all bibliographies associated with a story
    get_bibliographies_page_by_story_id(story_id: int, page: int, per_page: int)
        Get a single page of bibliographies associated with a story from the database
    search_bibliographies(search: str)
        Search for bibliographies by title associated with a user
    search_bibliographies_by_story_id(story_id: int, search: str)
        Search for bibliographies by title associated with a story
    """

    def __init__(
            self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_bibliography(
        self, story_id: int, title: str, pages: str = None,
        publication_date: str = None, other_text: str = None
    ) -> Bibliography:
        """Create a new bibliography

        Parameters
        ----------
        story_id : int
            The id of the story
        title : str
            The title of the referenced work
        pages : str
            The pages of the referenced work, optional
        publication_date : str
            The publication date of the referenced work, optional
        other_text : str
            Other text, optional

        Returns
        -------
        Bibliography
            The new bibliography object
        """

        with self._session as session:

            try:

                title_exists = session.query(Bibliography).filter(
                    Bibliography.title == title,
                    Bibliography.story_id == story_id,
                    Bibliography.user_id == self._owner.id
                ).first()

                if title_exists:
                    raise Exception('A reference with the same title is already \
                        associated with that story.')

                created = datetime.now()
                modified = created

                bibliography = Bibliography(
                    user_id=self._owner.id, story_id=story_id, title=title,
                    pages=pages, publication_date=publication_date,
                    other_text=other_text, created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Bibliography \
                    {bibliography.title[:50]} created by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(bibliography)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return bibliography

    def update_bibliography(
        self, bibliography_id: int, story_id: int, title: str,
        pages: str = None, publication_date: str = None, other_text: str = None
    ) -> Type[Bibliography]:
        """Update a bibliography

        Parameters
        ----------
        bibliography_id : int
            The id of the bibliography to update
        story_id : int
            The id of the story
        title : str
            The title of the referenced work
        pages : str
            The pages of the referenced work, optional
        publication_date : str
            The publication date of the referenced work, optional
        other_text : str
            Other text, optional

        Returns
        -------
        Bibliography
            The updated bibliography object
        """

        with self._session as session:

            try:

                bibliography = session.query(Bibliography).filter(
                    Bibliography.id == bibliography_id,
                    Bibliography.user_id == self._owner.id
                ).first()

                if not bibliography:
                    raise ValueError('Bibliography not found.')

                title_exists = session.query(Bibliography).filter(
                    Bibliography.title == title,
                    Bibliography.story_id == story_id
                ).first()

                if title_exists:
                    raise Exception('A reference with the same title is already\
                     associated with that story.')

                bibliography.story_id = story_id
                bibliography.title = title
                bibliography.pages = pages
                bibliography.publication_date = publication_date
                bibliography.other_text = other_text
                bibliography.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Bibliography \
                    {bibliography.title} updated by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return bibliography

    def delete_bibliography(self, bibliography_id: int) -> bool:
        """Delete a bibliography

        Before deleting the Bibliography, first delete any BibliographyAuthor objects associated with the Bibliography.

        Parameters
        ----------
        bibliography_id : int
            The id of the bibliography to delete

        Returns
        -------
        bool
            True if the bibliography was deleted, False if not
        """

        with self._session as session:

            try:

                bibliography = session.query(Bibliography).filter(
                    Bibliography.id == bibliography_id,
                    Bibliography.user_id == self._owner.id
                ).first()

                if not bibliography:
                    raise ValueError('Bibliography not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Bibliography \
                    {bibliography.title} deleted by {self._owner.username}',
                    created=datetime.now()
                )

                session.delete(bibliography)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_bibliography_by_id(
        self, bibliography_id: int
    ) -> Type[Bibliography] | None:
        """Get a bibliography by id

        Parameters
        ----------
        bibliography_id : int
            The id of the bibliography to get

        Returns
        -------
        Bibliography | None
            The bibliography object or None if not found
        """

        with self._session as session:

            bibliography = session.query(Bibliography).filter(
                Bibliography.id == bibliography_id,
                Bibliography.user_id == self._owner.id
            ).first()

            return bibliography if bibliography else None

    def get_bibliography_by_title(
        self, title: str
    ) -> Type[Bibliography] | None:
        """Get a bibliography by title

        Parameters
        ----------
        title : str
            The title of the referenced work

        Returns
        -------
        Bibliography | None
            The bibliography object or None if not found
        """

        with self._session as session:

            bibliography = session.query(Bibliography).filter(
                Bibliography.title == title,
                Bibliography.user_id == self._owner.id
            ).first()

            return bibliography if bibliography else None

    def get_bibliography_count(self) -> int:
        """Get bibliography count associated with a user

        Returns
        -------
        int
            The number of bibliographies
        """

        with self._session as session:

            return session.query(func.count(Bibliography.id)).filter(
                Bibliography.user_id == self._owner.id
            ).scalar()

    def get_all_bibliographies(self) -> List[Type[Bibliography]]:
        """Get all bibliographies associated with a user

        Returns
        -------
        list
            A list of bibliography objects
        """

        with self._session as session:

            return session.query(Bibliography).filter(
                Bibliography.user_id == self._owner.id
            ).all()

    def get_bibliographies_page(
        self, page: int, per_page: int
    ) -> List[Type[Bibliography]]:
        """Get a single page of bibliographies from the database associated with a user

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of bibliography objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Bibliography).filter(
                Bibliography.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def get_bibliographies_by_story_id(
        self, story_id: int
    ) -> List[Type[Bibliography]]:
        """Get all bibliographies associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of bibliography objects
        """

        with self._session as session:

            bibliographies = session.query(Bibliography).filter(
                Bibliography.story_id == story_id,
                Bibliography.user_id == self._owner.id
            ).all()

            return bibliographies if bibliographies else None

    def get_bibliographies_page_by_story_id(
        self, story_id: int, page: int, per_page: int
    ) -> List[Type[Bibliography]]:
        """Get a single page of bibliographies associated with a story from the database

        Parameters
        ----------
        story_id : int
            The id of the story
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of bibliography objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            bibliographies = session.query(Bibliography).filter(
                Bibliography.story_id == story_id,
                Bibliography.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

            return bibliographies if bibliographies else None

    def search_bibliographies(self, search: str) -> List[Type[Bibliography]]:
        """Search for bibliographies by title associated with a user

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of bibliography objects
        """

        with self._session as session:

            return session.query(Bibliography).filter(
                Bibliography.title.like(f'%{search}%'),
                Bibliography.user_id == self._owner.id
            ).all()

    def search_bibliographies_by_story_id(
        self, story_id: int, search: str
    ) -> List[Type[Bibliography]]:
        """Search for bibliographies by title associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story
        search : str
            The search string

        Returns
        -------
        list
            A list of bibliography objects
        """

        with self._session as session:

            return session.query(Bibliography).filter(
                Bibliography.story_id == story_id,
                Bibliography.title.like(f'%{search}%'),
                Bibliography.user_id == self._owner.id
            ).all()



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import ImageMimeTypes, User, CharacterImage, ImageLocation, Base


class Image(Base):
    """The Image class represents an image in the application.

    Attributes
    ----------
        id: int
            The image's id
        user_id: int
            The id of the owner of this entry
        caption: str
            The image's caption
        filename: str
            The image's filename
        dirname: str
            The image's directory name
        size_in_bytes: int
            The image's size in bytes
        mime_type: str
            The image's mime type
        created: str
            The creation datetime of the image
        modified: str
            The last modification datetime of the image
        user: User
            The user who owns this entry
        character: List[CharacterImage]
            The characters that the image has
        location: List[ImageLocation]
            The locations that the image has

    Methods
    -------
        __repr__()
            Returns a string representation of the image
        __str__()
            Returns a string representation of the image
        serialize()
            Returns a dictionary representation of the image
        unserialize(data: dict)
            Updates the image's attributes with the values from the dictionary
        validate_caption(caption: str)
            Validates the caption's length
        validate_filename(filename: str)
            Validates the filename's length
        validate_dirname(dirname: str)
            Validates the dirname's length
        validate_size_in_bytes(size_in_bytes: int)
            Validates the size in bytes
        validate_mime_type(mime_type: str)
            Validates the mime type
    """

    __tablename__ = 'images'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    filename: Mapped[str] = mapped_column(String(150), nullable=False)
    dirname: Mapped[str] = mapped_column(String(150), nullable=False)
    size_in_bytes: Mapped[int] = mapped_column(
        Integer, nullable=False, default=0
    )
    mime_type: Mapped[ImageMimeTypes] = mapped_column(
        String(50), nullable=False
    )
    caption: Mapped[str] = mapped_column(String(250), nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User", back_populates="images")
    character: Mapped[Optional[List["CharacterImage"]]] = relationship(
        "CharacterImage", back_populates="image",
        cascade="all, delete, delete-orphan")
    location: Mapped[Optional[List["ImageLocation"]]] = relationship(
        "ImageLocation", back_populates="image",
        cascade="all, delete, delete-orphan")

    def __repr__(self):
        """Returns a string representation of the image.

        Returns
        -------
        str
            A string representation of the image
        """

        return f'<Image {self.caption!r}>'

    def __str__(self):
        """Returns a string representation of the image.

        Returns
        -------
        str
            A string representation of the image
        """

        return f'{self.caption}' if self.caption else f'{self.filename}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the image.

        Returns
        -------
        dict
            A dictionary representation of the image
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'caption': self.caption,
            'filename': self.filename,
            'dirname': self.dirname,
            'size_in_bytes': self.size_in_bytes,
            'mime_type': self.mime_type,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Image":
        """Updates the image's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the image

        Returns
        -------
        Image
            The unserialized image
        """

        self.user_id = data.get('user_id', self.user_id)
        self.caption = data.get('caption', self.caption)
        self.filename = data.get('filename', self.filename)
        self.dirname = data.get('dirname', self.dirname)
        self.size_in_bytes = data.get('size_in_bytes', self.size_in_bytes)
        self.mime_type = data.get('mime_type', self.mime_type)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("caption")
    def validate_caption(self, key, caption: str) -> str:
        """Validates the caption's length.

        Parameters
        ----------
        caption: str
            The image's caption

        Returns
        -------
        str
            The validated caption
        """

        if caption and len(caption) > 250:
            raise ValueError("The image caption cannot have more than 250 characters.")
        return caption

    @validates("filename")
    def validate_filename(self, key, filename: str) -> str:
        """Validates the filename's length.

        Parameters
        ----------
        filename: str
            The image's filename

        Returns
        -------
        str
            The validated filename
        """

        if not filename:
            raise ValueError("A filename is required.")

        if len(filename) > 150:
            raise ValueError("The filename cannot have more than 150 characters.")

        return filename

    @validates("dirname")
    def validate_dirname(self, key, dirname: str) -> str:
        """Validates the dirname's length.

        Parameters
        ----------
        dirname: str
            The image's directory name

        Returns
        -------
        str
            The validated directory name
        """

        if not dirname:
            raise ValueError("A directory name is required.")

        if len(dirname) > 150:
            raise ValueError("The directory name cannot have more than 150 characters.")

        return dirname

    @validates("size_in_bytes")
    def validate_size_in_bytes(self, key, size_in_bytes: int) -> int:
        """Validates the size in bytes.

        Parameters
        ----------
        size_in_bytes: int
            The image's size in bytes

        Returns
        -------
        int
            The validated size in bytes
        """

        if size_in_bytes < 0:
            raise ValueError("The size in bytes cannot be negative.")

        return size_in_bytes

    @validates("mime_type")
    def validate_mime_type(self, key, mime_type: str) -> str:
        """Validates the mime type.

        Parameters
        ----------
        mime_type: str
            The image's mime type

        Returns
        -------
        str
            The validated mime type
        """

        if mime_type not in [e.value for e in ImageMimeTypes]:
            raise ValueError("Invalid mime type.")

        return mime_type



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, String, Boolean, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import Activity, Assistance, Author, Character, Event, Image, Link, Location, Note, Story, \
    Submission, Base
from validators import email as email_validator
from validators import uuid as uuid_validator


class User(Base):
    """The User class represents a user in the application or the system user.

    Attributes
    ----------
        id: int
            The user's id
        uuid: str
            The user's UUID
        username: str
            The user's username
        email: str
            The user's email address
        password: str
            The user's password hash
        is_active: bool
            The user's active status
        is_banned: bool
            The user's banned status
        created: str
            The creation datetime of the user
        modified: str
            The last modification datetime of the user
        activities: List[Activity]
            The activities that the user has
        authors: List[Author]
            The authors that the user has
        characters: List[Character]
            The characters that the user has
        events: List[Event]
            The events that the user has
        images: List[Image]
            The images that the user has
        links: List[Link]
            The links that the user has
        locations: List[Location]
            The locations that the user has
        notes: List[Note]
            The notes that the user has
        stories: List[Story]
            The stories that the user has
        submissions: List[Submission]
            The submissions that the user has

    Methods
    -------
        __repr__()
            Returns a string representation of the user
        __str__()
            Returns a string representation of the user
        serialize()
            Returns a dictionary representation of the user
        unserialize(data: dict)
            Updates the user's attributes with the values from the dictionary
        validate_uuid(uuid: str)
            Validates the UUID's length and format
        validate_username(username: str)
            Validates the username's length
        validate_email(email: str)
            Validates the email's length
        validate_password(password: str)
            Validates the password's length
    """

    __tablename__ = 'users'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    uuid: Mapped[str] = mapped_column(String(36), nullable=False)
    username: Mapped[str] = mapped_column(String(50), nullable=False)
    email: Mapped[str] = mapped_column(String(100), nullable=False)
    password: Mapped[str] = mapped_column(String(250), nullable=False)
    is_active: Mapped[bool] = mapped_column(Boolean, default=False)
    is_banned: Mapped[bool] = mapped_column(Boolean, default=False)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    activities: Mapped[Optional[List["Activity"]]] = relationship(
        "Activity", back_populates="user",
        cascade="all, delete, delete-orphan")
    assistances: Mapped[Optional[List["Assistance"]]] = relationship(
        "Assistance", back_populates="user",
        cascade="all, delete, delete-orphan")
    authors: Mapped[Optional[List["Author"]]] = relationship(
        "Author", back_populates="user", lazy="joined",
        cascade="all, delete, delete-orphan")
    characters: Mapped[Optional[List["Character"]]] = relationship(
        "Character", back_populates="user",
        cascade="all, delete, delete-orphan")
    events: Mapped[Optional[List["Event"]]] = relationship(
        "Event", back_populates="user",
        cascade="all, delete, delete-orphan")
    images: Mapped[Optional[List["Image"]]] = relationship(
        "Image", back_populates="user",
        cascade="all, delete, delete-orphan")
    links: Mapped[Optional[List["Link"]]] = relationship(
        "Link", back_populates="user",
        cascade="all, delete, delete-orphan")
    locations: Mapped[Optional[List["Location"]]] = relationship(
        "Location", back_populates="user",
        cascade="all, delete, delete-orphan")
    notes: Mapped[Optional[List["Note"]]] = relationship(
        "Note", back_populates="user",
        cascade="all, delete, delete-orphan")
    stories: Mapped[Optional[List["Story"]]] = relationship(
        "Story", back_populates="user",
        cascade="all, delete, delete-orphan")
    submissions: Mapped[Optional[List["Submission"]]] = relationship(
        "Submission", back_populates="user",
        cascade="all, delete, delete-orphan")

    def __repr__(self):
        """Returns a string representation of the user.

        Returns
        -------
        str
            A string representation of the user
        """

        return f'<User {self.id!r} - {self.username!r}>'

    def __str__(self):
        """Returns a string representation of the user.

        Returns
        -------
        str
            A string representation of the user
        """

        return f'{self.id!r} - {self.username}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the user.

        Returns
        -------
        dict
            A dictionary representation of the user
        """

        return {
            'id': self.id,
            'uuid': self.uuid,
            'username': self.username,
            'password': self.password,
            'email': self.email,
            'is_active': self.is_active,
            'is_banned': self.is_banned,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "User":
        """Updates the user's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the user

        Returns
        -------
        User
            The unserialized user
        """

        self.uuid = data.get('uuid', self.uuid)
        self.username = data.get('username', self.username)
        self.email = data.get('email', self.email)
        self.is_active = data.get('is_active', self.is_active)
        self.is_banned = data.get('is_banned', self.is_banned)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("uuid")
    def validate_uuid(self, key, uuid: str) -> str:
        """Validates the UUID's length and format.

        Parameters
        ----------
        uuid: str
            The user's UUID

        Returns
        -------
        str
            The validated UUID
        """

        if not uuid:
            raise ValueError("A user UUID is required.")

        if len(uuid) != 36:
            raise ValueError("The user UUID must have 36 characters.")

        if not uuid_validator(uuid):
            raise ValueError("The user UUID is not valid.")

        return uuid

    @validates("username")
    def validate_username(self, key, username: str) -> str:
        """Validates the username's length.

        Parameters
        ----------
        username: str
            The user's username

        Returns
        -------
        str
            The validated username
        """

        if not username:
            raise ValueError("A username is required.")

        if len(username) > 50:
            raise ValueError("The username can have no more than 50 characters.")

        return username

    @validates("email")
    def validate_email(self, key, email: str) -> str:
        """Validates the email's length.

        Parameters
        ----------
        email: str
            The user's email address

        Returns
        -------
        str
            The validated email
        """

        if not email:
            raise ValueError("An email address is required.")

        if len(email) > 100:
            raise ValueError("The email address can have no more than 100 characters.")

        if not email_validator(email):
            raise ValueError("The email address is not valid.")

        return email

    @validates("password")
    def validate_password(self, key, password: str) -> str:
        """Validates the password's length.

        Parameters
        ----------
        password: str
            The user's password hash

        Returns
        -------
        str
            The validated password
        """

        if not password:
            raise ValueError("A password is required.")

        if len(password) > 250:
            raise ValueError("The password can have no more than 250 characters.")

        return password



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Event, Link, Base


class EventLink(Base):
    """The EventLink class represents the relationship between an event and a link.

    Attributes
    ----------
        user_id: int
            The id of the owner of this entry
        event_id: int
            The event's id
        link_id: int
            The link's id
        created: str
            The creation datetime of the link between the Event and the Link
        user: User
            The user who owns this entry
        event: Event
            The event that the link belongs to
        link: Link
            The link that the event has

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """

    __tablename__ = 'events_links'
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    event_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('events.id'), primary_key=True
    )
    link_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('links.id'), primary_key=True
    )
    created: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User")
    event: Mapped["Event"] = relationship("Event", back_populates="links")
    link: Mapped["Link"] = relationship("Link", back_populates="events")

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<EventLink {self.event.title!r} - {self.link.title!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'{self.event.title} - {self.link.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'user_id': self.user_id,
            'event_id': self.event_id,
            'link_id': self.link_id,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "EventLink":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        EventLink
            The unserialized relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.event_id = data.get('event_id', self.event_id)
        self.link_id = data.get('link_id', self.link_id)
        self.created = data.get('created', self.created)

        return self



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, Text, Float, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, ImageLocation, LinkLocation, LocationNote, Base


class Location(Base):
    """The Location class represents a location associated with one or more stories.

    Attributes
    ----------
        id: int
            The location's id
        user_id: int
            The id of the owner of this entry
        title: str
            The location's title
        description: str
            The location's description
        address: str
            The location's address
        city: str
            The location's city
        state: str
            The location's state
        country: str
            The location's country
        zip_code: str
            The location's zip code
        latitude: float
            The location's latitude
        longitude: float
            The location's longitude
        created: str
            The creation datetime of the location
        modified: str
            The last modification datetime of the location
        user: User
            The user who owns this entry
        images: List[ImageLocation]
            The images that the location has
        links: List[LinkLocation]
            The links that the location has
        notes: List[LocationNote]
            The notes that the location has

    Methods
    -------
        __repr__()
            Returns a string representation of the location
        __str__()
            Returns a string representation of the location
        serialize()
            Returns a dictionary representation of the location
        unserialize(data: dict)
            Updates the location's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_description(description: str)
            Validates the description's length
        validate_address(address: str)
            Validates the address's length
        validate_city(city: str)
            Validates the city's length
        validate_state(state: str)
            Validates the state's length
        validate_country(country: str)
            Validates the country's length
        validate_zip_code(zip_code: str)
            Validates the zip code's length
        validate_latitude(latitude: float)
            Validates the latitude's range
        validate_longitude(longitude: float)
            Validates the longitude's range
    """

    __tablename__ = 'locations'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    title: Mapped[str] = mapped_column(String(250), nullable=True)
    description: Mapped[str] = mapped_column(Text, nullable=True)
    address: Mapped[str] = mapped_column(String(250), nullable=True)
    city: Mapped[str] = mapped_column(String(100), nullable=True)
    state: Mapped[str] = mapped_column(String(100), nullable=True)
    country: Mapped[str] = mapped_column(String(100), nullable=True)
    zip_code: Mapped[str] = mapped_column(String(20), nullable=True)
    latitude: Mapped[float] = mapped_column(Float, nullable=True)
    longitude: Mapped[float] = mapped_column(Float, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User", back_populates="locations")
    images: Mapped[Optional[List["ImageLocation"]]] = relationship(
        "ImageLocation", back_populates="location",
        cascade="all, delete, delete-orphan")
    links: Mapped[Optional[List["LinkLocation"]]] = relationship(
        "LinkLocation", back_populates="location",
        cascade="all, delete, delete-orphan", lazy="joined")
    notes: Mapped[Optional[List["LocationNote"]]] = relationship(
        "LocationNote", back_populates="location",
        cascade="all, delete, delete-orphan", lazy="joined")

    def __repr__(self):
        """Returns a string representation of the location.

        Returns
        -------
        str
            A string representation of the location
        """

        return f'<Location {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the location.

        Returns
        -------
        str
            A string representation of the location
        """

        return f'{self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the location.

        Returns
        -------
        dict
            A dictionary representation of the location
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'title': self.title,
            'description': self.description,
            'address': self.address,
            'city': self.city,
            'state': self.state,
            'country': self.country,
            'zip_code': self.zip_code,
            'latitude': self.latitude,
            'longitude': self.longitude,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Location":
        """Updates the location's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the location

        Returns
        -------
        Location
            The unserialized location
        """

        self.user_id = data.get('user_id', self.user_id)
        self.title = data.get('title', self.title)
        self.description = data.get('description', self.description)
        self.address = data.get('address', self.address)
        self.city = data.get('city', self.city)
        self.state = data.get('state', self.state)
        self.country = data.get('country', self.country)
        self.zip_code = data.get('zip_code', self.zip_code)
        self.latitude = data.get('latitude', self.latitude)
        self.longitude = data.get('longitude', self.longitude)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length

        Parameters
        ----------
        title: str
            The location's title

        Returns
        -------
        str
            The validated title
        """

        if title and len(title) > 250:
            raise ValueError("The location title can have no more than 250 characters.")

        return title

    @validates("description")
    def validate_description(self, key, description: str) -> str:
        """Validates the description's length

        Parameters
        ----------
        description: str
            The location's description

        Returns
        -------
        str
            The validated description
        """

        if description and len(description) > 65535:
            raise ValueError("The location description can have no more than 65,535 characters.")

        return description

    @validates("address")
    def validate_address(self, key, address: str) -> str:
        """Validates the address's length

        Parameters
        ----------
        address: str
            The location's address

        Returns
        -------
        str
            The validated address
        """

        if address and len(address) > 250:
            raise ValueError("The location address can have no more than 250 characters.")

        return address

    @validates("city")
    def validate_city(self, key, city: str) -> str:
        """Validates the city's length

        Parameters
        ----------
        city: str
            The location's city

        Returns
        -------
        str
            The validated city
        """

        if city and len(city) > 100:
            raise ValueError("The location city can have no more than 100 characters.")

        return city

    @validates("state")
    def validate_state(self, key, state: str) -> str:
        """Validates the state's length

        Parameters
        ----------
        state: str
            The location's state

        Returns
        -------
        str
            The validated state
        """

        if state and len(state) > 100:
            raise ValueError("The location state can have no more than 100 characters.")

        return state

    @validates("country")
    def validate_country(self, key, country: str) -> str:
        """Validates the country's length

        Parameters
        ----------
        country: str
            The location's country

        Returns
        -------
        str
            The validated country
        """

        if country and len(country) > 100:
            raise ValueError("The location country can have no more than 100 characters.")

        return country

    @validates("zip_code")
    def validate_zip_code(self, key, zip_code: str) -> str:
        """Validates the zip code's length

        Parameters
        ----------
        zip_code: str
            The location's zip code

        Returns
        -------
        str
            The validated zip code
        """

        if zip_code and len(zip_code) > 20:
            raise ValueError("The location zip code can have no more than 20 characters.")

        return zip_code

    @validates("latitude")
    def validate_latitude(self, key, latitude: float) -> float:
        """Validates the latitude's range

        Parameters
        ----------
        latitude: float
            The location's latitude

        Returns
        -------
        float
            The validated latitude
        """

        if type(latitude) is float:
            if latitude < -90.0 or latitude > 90.0:
                raise ValueError("The location latitude must be between -90 and 90.")

        return latitude

    @validates("longitude")
    def validate_longitude(self, key, longitude: float) -> float:
        """Validates the longitude's range

        Parameters
        ----------
        longitude: float
            The location's longitude

        Returns
        -------
        float
            The validated longitude
        """

        if type(longitude) is float:
            if longitude < -180.0 or longitude > 180.0:
                raise ValueError("The location longitude must be between -180 and 180.")

        return longitude



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, Boolean, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Image, Base, Character


class CharacterImage(Base):
    """The CharacterImage class represents the relationship between a character and an image.

    Attributes
    ----------
        user_id: int
            The id of the owner of this entry
        character_id: int
            The character's id
        image_id: int
            The image's id
        position: int
            The position of the image in the character's gallery
        is_default: bool
            Whether the image is the default image for the character
        created: str
            The creation datetime of the link between the Character and the Image
        modified: str
            The last modification datetime of the link between the Character and the Image
        user: User
            The user who owns this entry
        character: Character
            The character that the image belongs to
        image: Image
            The image that the character has

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """
    __tablename__ = 'characters_images'
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    character_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('characters.id'), primary_key=True
    )
    image_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('images.id'), primary_key=True
    )
    position: Mapped[int] = mapped_column(Integer, nullable=False, default=1)
    is_default: Mapped[bool] = mapped_column(Boolean, default=False)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User")
    character: Mapped["Character"] = relationship(
        "Character", back_populates="images"
    )
    image: Mapped["Image"] = relationship(
        "Image", back_populates="character", lazy="joined"
    )

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<CharacterImage {self.character.first_name!r} {self.character.last_name!r} - {self.image.caption!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'{self.character.first_name} {self.character.last_name} - {self.image.caption}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'user_id': self.user_id,
            'character_id': self.character_id,
            'image_id': self.image_id,
            'position': self.position,
            'is_default': self.is_default,
            'created': str(self.created),
            'modified': str(self.modified),
            'image': self.image.serialize()
        }

    def unserialize(self, data: dict) -> "CharacterImage":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        CharacterImage
            The updated relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.character_id = data.get('character_id', self.character_id)
        self.image_id = data.get('image_id', self.image_id)
        self.position = data.get('position', self.position)
        self.is_default = data.get('is_default', self.is_default)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Character, Note, Base


class CharacterNote(Base):
    """The CharacterNote class represents the relationship between a character and a note.

    Attributes
    ----------
        user_id: int
            The id of the owner of this entry
        character_id: int
            The character's id
        note_id: int
            The note's id
        created: str
            The creation datetime of the link between the Character and the Note
        user: User
            The user who owns this entry
        character: Character
            The character that the note belongs to
        note: Note
            The note that the character has

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """

    __tablename__ = 'characters_notes'
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    character_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('characters.id'), primary_key=True
    )
    note_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('notes.id'), primary_key=True
    )
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship("User")
    character: Mapped["Character"] = relationship(
        "Character", back_populates="notes"
    )
    note: Mapped["Note"] = relationship(
        "Note", back_populates="characters", lazy='joined'
    )

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<CharacterNote {self.character.first_name!r} {self.character.last_name!r} - {self.note.title!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'{self.character.first_name} {self.character.last_name} - {self.note.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'user_id': self.user_id,
            'character_id': self.character_id,
            'note_id': self.note_id,
            'created': str(self.created),
            'note': self.note.serialize()
        }

    def unserialize(self, data: dict) -> "CharacterNote":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        CharacterNote
            The updated relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.character_id = data.get('character_id', self.character_id)
        self.note_id = data.get('note_id', self.note_id)
        self.created = data.get('created', self.created)

        return self



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, Text, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, Chapter, AuthorStory, Bibliography, Submission, LinkStory, NoteStory, CharacterStory, \
    Base


class Story(Base):
    """The Story class represents a story in the application.

    Attributes
    ----------
        id: int
            The story's id
        user_id: int
            The id of the owner of this entry
        title: str
            The story's title
        description: str
            The story's description
        created: str
            The creation datetime of the story
        modified: str
            The last modification datetime of the story
        user: User
            The user who owns this entry
        chapters: List[Chapter]
            The chapters that the story has
        authors: List[AuthorStory]
            The authors that the story has
        references: List[Bibliography]
            The references that the story has
        submissions: List[Submission]
            The submissions that the story has
        links: List[LinkStory]
            The links that the story has
        notes: List[NoteStory]
            The notes that the story has

    Methods
    -------
        __repr__()
            Returns a string representation of the story
        __str__()
            Returns a string representation of the story
        serialize()
            Returns a dictionary representation of the story
        unserialize(data: dict)
            Updates the story's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_description(description: str)
            Validates the description's length
    """

    __tablename__ = 'stories'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    title: Mapped[str] = mapped_column(String(250), nullable=False)
    description: Mapped[str] = mapped_column(Text, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User", back_populates="stories")
    chapters: Mapped[Optional[List["Chapter"]]] = relationship(
        "Chapter", back_populates="story", lazy="joined",
        cascade="all, delete, delete-orphan")
    authors: Mapped[Optional[List["AuthorStory"]]] = relationship(
        "AuthorStory", back_populates="story", lazy="joined",
        cascade="all, delete, delete-orphan")
    references: Mapped[Optional[List["Bibliography"]]] = relationship(
        "Bibliography", back_populates="story", lazy="joined",
        cascade="all, delete, delete-orphan")
    submissions: Mapped[Optional[List["Submission"]]] = relationship(
        "Submission", back_populates="story", lazy="joined",
        cascade="all, delete, delete-orphan")
    links: Mapped[Optional[List["LinkStory"]]] = relationship(
        "LinkStory", back_populates="story", lazy="joined",
        cascade="all, delete, delete-orphan")
    notes: Mapped[Optional[List["NoteStory"]]] = relationship(
        "NoteStory", back_populates="story", lazy="joined",
        cascade="all, delete, delete-orphan")
    characters: Mapped[Optional[List["CharacterStory"]]] = relationship(
        "CharacterStory", back_populates="story",
        cascade="all, delete, delete-orphan")

    def __repr__(self):
        """Returns a string representation of the story.

        Returns
        -------
        str
            A string representation of the story
        """

        return f'<Story {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the story.

        Returns
        -------
        str
            A string representation of the story
        """

        return f'{self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the story.

        Returns
        -------
        dict
            A dictionary representation of the story
        """

        authors = []
        if self.authors:
            for author_story in self.authors:
                authors.append(author_story.author.serialize())

        links = []
        if self.links:
            for link_story in self.links:
                links.append(link_story.link.serialize())

        notes = []
        if self.notes:
            for note_story in self.notes:
                notes.append(note_story.note.serialize())

        chapters = []
        if self.chapters:
            for chapter in self.chapters:
                chapters.append(chapter.serialize())

        escaped_title = self.title.replace('"', "'").replace('/', '\\/')
        escaped_description = None

        if self.description:
            escaped_description = self.description.replace('"', '\\"').replace('/', '\\/')

        return {
            'id': self.id,
            'user_id': self.user_id,
            'title': escaped_title,
            'description': escaped_description,
            'created': str(self.created),
            'modified': str(self.modified),
            'authors': authors,
            'links': links,
            'notes': notes,
            'chapters': chapters
        }

    def unserialize(self, data: dict) -> "Story":
        """Updates the story's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the story

        Returns
        -------
        Story
            The unserialized story
        """

        self.user_id = data.get('user_id', self.user_id)
        self.title = data.get('title', self.title)
        self.description = data.get('description', self.description)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length.

        Parameters
        ----------
        title: str
            The story's title

        Returns
        -------
        str
            The validated title
            :param key:
        """

        if not title:
            raise ValueError("A story title is required.")

        if len(title) > 250:
            raise ValueError("The story title can have no more than 250 characters.")

        return title

    @validates("description")
    def validate_description(self, key, description: str) -> str:
        """Validates the description's length.

        Parameters
        ----------
        description: str
            The story's description

        Returns
        -------
        str
            The validated description
        """

        if description and len(description) > 65535:
            raise ValueError("The story description can have no more than 65,535 characters.")

        return description



from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, DateTime, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, EventLink, CharacterEvent, EventNote, Base


class Event(Base):
    """The Event class represents an event that is referenced by one or more stories.

    Attributes
    ----------
        id: int
            The event's id
        user_id: int
            The id of the owner of this entry
        title: str
            The event's title
        description: str
            A description of the event
        start_datetime: str
            The starting datetime of the event
        end_datetime: str
            The ending datetime of the event
        created: str
            The creation datetime of the event
        modified: str
            The last modification datetime of the event
        user: User
            The user who owns this entry
        links: List[EventLink]
            The links that the event has
        characters: List[CharacterEvent]
            The characters that the event has
        notes: List[EventNote]
            The notes that the event has

    Methods
    -------
        __repr__()
            Returns a string representation of the event
        __str__()
            Returns a string representation of the event
        serialize()
            Returns a dictionary representation of the event
        unserialize(data: dict)
            Updates the event's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_description(description: str)
            Validates the description's length
        validate_start_datetime(start_datetime: str)
            Validates the start datetime's format
        validate_end_datetime(end_datetime: str)
            Validates the end datetime's format
    """

    __tablename__ = 'events'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    title: Mapped[str] = mapped_column(String(250), nullable=False)
    description: Mapped[str] = mapped_column(Text, nullable=True)
    start_datetime: Mapped[str] = mapped_column(DateTime, nullable=True)
    end_datetime: Mapped[str] = mapped_column(DateTime, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship(
        "User", back_populates="events"
    )
    links: Mapped[Optional[List["EventLink"]]] = relationship(
        "EventLink", back_populates="event", lazy="joined",
        cascade="all, delete, delete-orphan")
    characters: Mapped[Optional[List["CharacterEvent"]]] = relationship(
        "CharacterEvent", back_populates="event", lazy="joined",
        cascade="all, delete, delete-orphan")
    notes: Mapped[Optional[List["EventNote"]]] = relationship(
        "EventNote", back_populates="event", lazy="joined",
        cascade="all, delete, delete-orphan")

    def __repr__(self):
        """Returns a string representation of the event.

        Returns
        -------
        str
            A string representation of the event
        """

        return f'<Event {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the event.

        Returns
        -------
        str
            A string representation of the event
        """

        return f'{self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the event.

        Returns
        -------
        dict
            A dictionary representation of the event
        """

        return {
            'user_id': self.user_id,
            'title': self.title,
            'description': self.description,
            'start_datetime': self.start_datetime,
            'end_datetime': self.end_datetime,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Event":
        """Updates the event's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the event

        Returns
        -------
        Event
            The unserialized event
        """

        self.user_id = data.get('user_id', self.user_id)
        self.title = data.get('title', self.title)
        self.description = data.get('description', self.description)
        self.start_datetime = data.get('start_datetime', self.start_datetime)
        self.end_datetime = data.get('end_datetime', self.end_datetime)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length.

        Parameters
        ----------
        title: str
            The event's title

        Returns
        -------
        str
            The validated title
        """

        if not title:
            raise ValueError("Title cannot be empty")

        return title

    @validates("description")
    def validate_description(self, key, description: str) -> str:
        """Validates the description's length.

        Parameters
        ----------
        description: str
            The event's description

        Returns
        -------
        str
            The validated description
        """

        if description and len(description) > 65535:
            raise ValueError("Description cannot have more than 65,535 characters.")

        return description



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, String, Text, Float, Boolean, BigInteger, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Base


class Assistance(Base):
    """The Assistance class represents an assistance request.

    Attributes
    ----------
    id: int
        The Assistance id
    user_id: int
        The id of the current user
    session_uuid: str
        The UUID of the session to be used when making the request. If other
        messages in the database have the same session_uuid value, they will be
        loaded in order of their creation date, and the assistant will continue
        from the last message in the session.
    assistant: str
        The assistant to be used when making the request.
    model: str
        The language model to be used when making the request.
    priming: str
        The priming to be used when making the request.
    prompt: str
        The prompt to be used when making the request.
    temperature: float
        The temperature to be used when making the request.
    seed: int
        The seed to be used when making the request.
    content: str
        The language model's response to the prompt.
    done: bool
        Whether the assistance is done or not
    total_duration: int
        The total duration of the assistance (ns)
    load_duration: int
        The duration of the loading of the assistance (ns)
    prompt_eval_count: int
        The number of prompt evaluations
    prompt_eval_duration: int
        The duration of the prompt evaluations (ns)
    eval_count: int
        The number of evaluations
    eval_duration: int
        The duration of the evaluations (ns)
    created: str
        The assistance's creation date in datetime form: yyy-mm-dd hh:mm:ss.f
    """

    __tablename__ = 'assistances'
    id: Mapped[int] = mapped_column(
        Integer, primary_key=True, autoincrement=True
    )
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    session_uuid: Mapped[str] = mapped_column(String(36), nullable=False)
    assistant: Mapped[str] = mapped_column(String(100), nullable=False)
    model: Mapped[str] = mapped_column(String(100), nullable=False)
    priming: Mapped[str] = mapped_column(Text, nullable=True)
    prompt: Mapped[str] = mapped_column(Text, nullable=True)
    temperature: Mapped[float] = mapped_column(
        Float, nullable=False, default=0.5
    )
    seed: Mapped[int] = mapped_column(Integer, nullable=True)
    content: Mapped[str] = mapped_column(Text, nullable=True)
    done: Mapped[bool] = mapped_column(Boolean, default=False)
    total_duration: Mapped[int] = mapped_column(
        BigInteger, nullable=True, default=0
    )
    load_duration: Mapped[int] = mapped_column(
        BigInteger, nullable=True, default=0
    )
    prompt_eval_count: Mapped[int] = mapped_column(
        BigInteger, nullable=True, default=0,
    )
    prompt_eval_duration: Mapped[int] = mapped_column(
        BigInteger, nullable=True, default=0
    )
    eval_count: Mapped[int] = mapped_column(
        BigInteger, nullable=True, default=0
    )
    eval_duration: Mapped[int] = mapped_column(
        BigInteger, nullable=True, default=0
    )
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship(
        "User", back_populates="assistances"
    )

    def __repr__(self):
        """Returns a string representation of the assistance.

        Returns
        -------
        str
            A string representation of the assistance
        """

        return f'<Assistant {self.assistant!r}>'

    def __str__(self):
        """Returns a string representation of the assistance.

        Returns
        -------
        str
            A string representation of the assistance
        """

        return f'Assistant: {self.assistant}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the assistance.

        Returns
        -------
        dict
            A dictionary representation of the assistance
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'session_uuid': self.session_uuid,
            'assistant': self.assistant,
            'model': self.model,
            'priming': self.priming,
            'prompt': self.prompt,
            'temperature': self.temperature,
            'seed': self.seed,
            'content': self.content,
            'done': self.done,
            'total_duration': self.total_duration,
            'load_duration': self.load_duration,
            'prompt_eval_count': self.prompt_eval_count,
            'prompt_eval_duration': self.prompt_eval_duration,
            'eval_count': self.eval_count,
            'eval_duration': self.eval_duration,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "Assistance":
        """Updates the assistance's attributes with the values from the dictionary.

        Returns
        -------
        Assistance
            The updated assistance
        """

        self.id = data.get('id', self.id)
        self.user_id = data.get('user_id', self.user_id)
        self.session_uuid = data.get('session_uuid', self.session_uuid)
        self.assistant = data.get('assistant', self.assistant)
        self.model = data.get('model', self.model)
        self.priming = data.get('priming', self.priming)
        self.prompt = data.get('prompt', self.prompt)
        self.temperature = data.get('temperature', self.temperature)
        self.seed = data.get('seed', self.seed)
        self.content = data.get('content', self.content)
        self.done = data.get('done', self.done)
        self.total_duration = data.get('total_duration', self.total_duration)
        self.load_duration = data.get('load_duration', self.load_duration)
        self.prompt_eval_count = data.get('prompt_eval_count', self.prompt_eval_count)
        self.prompt_eval_duration = data.get('prompt_eval_duration', self.prompt_eval_duration)
        self.eval_count = data.get('eval_count', self.eval_count)
        self.eval_duration = data.get('eval_duration', self.eval_duration)
        self.created = data.get('created', self.created)

        return self
